#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np

from pymc3.step_methods.arraystep import Competence
from pymc3.step_methods.hmc.base_hmc import BaseHMC, DivergenceInfo, HMCStepData
from pymc3.step_methods.hmc.integration import IntegrationError
from pymc3.vartypes import discrete_types

__all__ = ["HamiltonianMC"]


def unif(step_size, elow=0.85, ehigh=1.15):
    return np.random.uniform(elow, ehigh) * step_size


class HamiltonianMC(BaseHMC):
    R"""A sampler for continuous variables based on Hamiltonian mechanics.

    See NUTS sampler for automatically tuned stopping time and step size scaling.
    """

    name = "hmc"
    default_blocked = True
    generates_stats = True
    stats_dtypes = [
        {
            "step_size": np.float64,
            "n_steps": np.int64,
            "tune": bool,
            "step_size_bar": np.float64,
            "accept": np.float64,
            "diverging": bool,
            "energy_error": np.float64,
            "energy": np.float64,
            "path_length": np.float64,
            "accepted": bool,
            "model_logp": np.float64,
            "process_time_diff": np.float64,
            "perf_counter_diff": np.float64,
            "perf_counter_start": np.float64,
        }
    ]

    def __init__(self, vars=None, path_length=2.0, max_steps=1024, **kwargs):
        """Set up the Hamiltonian Monte Carlo sampler.

        Parameters
        ----------
        vars: list of theano variables
        path_length: float, default=2
            total length to travel
        step_rand: function float -> float, default=unif
            A function which takes the step size and returns an new one used to
            randomize the step size at each iteration.
        step_scale: float, default=0.25
            Initial size of steps to take, automatically scaled down
            by 1/n**(1/4).
        scaling: array_like, ndim = {1,2}
            The inverse mass, or precision matrix. One dimensional arrays are
            interpreted as diagonal matrices. If `is_cov` is set to True,
            this will be interpreded as the mass or covariance matrix.
        is_cov: bool, default=False
            Treat the scaling as mass or covariance matrix.
        potential: Potential, optional
            An object that represents the Hamiltonian with methods `velocity`,
            `energy`, and `random` methods. It can be specified instead
            of the scaling matrix.
        target_accept: float, default 0.65
            Adapt the step size such that the average acceptance
            probability across the trajectories are close to target_accept.
            Higher values for target_accept lead to smaller step sizes.
            Setting this to higher values like 0.9 or 0.99 can help
            with sampling from difficult posteriors. Valid values are
            between 0 and 1 (exclusive). Default of 0.65 is from (Beskos et.
            al. 2010, Neal 2011). See Hoffman and Gelman's "The No-U-Turn
            Sampler: Adaptively Setting Path Lengths in Hamiltonian Monte
            Carlo" section 3.2 for details.
        gamma: float, default .05
        k: float, default .75
            Parameter for dual averaging for step size adaptation. Values
            between 0.5 and 1 (exclusive) are admissible. Higher values
            correspond to slower adaptation.
        t0: float > 0, default 10
            Parameter for dual averaging. Higher values slow initial
            adaptation.
        adapt_step_size: bool, default=True
            Whether step size adaptation should be enabled. If this is
            disabled, `k`, `t0`, `gamma` and `target_accept` are ignored.
        max_steps: int
            The maximum number of leapfrog steps.
        model: pymc3.Model
            The model
        **kwargs: passed to BaseHMC
        """
        kwargs.setdefault("step_rand", unif)
        kwargs.setdefault("target_accept", 0.65)
        super().__init__(vars, **kwargs)
        self.path_length = path_length
        self.max_steps = max_steps

    def _hamiltonian_step(self, start, p0, step_size):
        n_steps = max(1, int(self.path_length / step_size))
        n_steps = min(self.max_steps, n_steps)

        energy_change = -np.inf
        state = start
        last = state
        div_info = None
        try:
            for _ in range(n_steps):
                last = state
                state = self.integrator.step(step_size, state)
        except IntegrationError as e:
            div_info = DivergenceInfo("Integration failed.", e, last, None)
        else:
            if not np.isfinite(state.energy):
                div_info = DivergenceInfo("Divergence encountered, bad energy.", None, last, state)
            energy_change = start.energy - state.energy
            if np.isnan(energy_change):
                energy_change = -np.inf
            if np.abs(energy_change) > self.Emax:
                div_info = DivergenceInfo(
                    "Divergence encountered, large integration error.", None, last, state
                )

        accept_stat = min(1, np.exp(energy_change))

        if div_info is not None or np.random.rand() >= accept_stat:
            end = start
            accepted = False
        else:
            end = state
            accepted = True

        stats = {
            "path_length": self.path_length,
            "n_steps": n_steps,
            "accept": accept_stat,
            "energy_error": energy_change,
            "energy": state.energy,
            "accepted": accepted,
            "model_logp": state.model_logp,
        }
        return HMCStepData(end, accept_stat, div_info, stats)

    @staticmethod
    def competence(var, has_grad):
        """Check how appropriate this class is for sampling a random variable."""
        if var.dtype in discrete_types or not has_grad:
            return Competence.INCOMPATIBLE
        return Competence.COMPATIBLE



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy.testing as npt

from pymc3.backends import ndarray
from pymc3.backends import tracetab as ttab
from pymc3.tests import backend_fixtures as bf


class TestTraceToDf(bf.ModelBackendSampledTestCase):
    backend = ndarray.NDArray
    name = "text-db"
    shape = (2, 3)

    def test_trace_to_dataframe(self):
        mtrace = self.mtrace
        df = ttab.trace_to_dataframe(mtrace)
        assert len(mtrace) * mtrace.nchains == df.shape[0]

        checked = False
        for varname in self.test_point.keys():
            vararr = mtrace.get_values(varname)
            # With `shape` above, only one variable has to have that
            # `shape`.
            if vararr.shape[1:] != self.shape:
                continue
            npt.assert_equal(vararr[:, 0, 0], df[varname + "__0_0"].values)
            npt.assert_equal(vararr[:, 1, 0], df[varname + "__1_0"].values)
            npt.assert_equal(vararr[:, 1, 2], df[varname + "__1_2"].values)
            checked = True
        assert checked

    def test_trace_to_dataframe_chain_arg(self):
        mtrace = self.mtrace
        df = ttab.trace_to_dataframe(mtrace, chains=0)
        assert len(mtrace) == df.shape[0]

        checked = False
        for varname in self.test_point.keys():
            vararr = mtrace.get_values(varname, chains=0)
            # With `shape` above, only one variable has to have that
            # `shape`.
            if vararr.shape[1:] != self.shape:
                continue
            npt.assert_equal(vararr[:, 0, 0], df[varname + "__0_0"].values)
            npt.assert_equal(vararr[:, 1, 0], df[varname + "__1_0"].values)
            npt.assert_equal(vararr[:, 1, 2], df[varname + "__1_2"].values)
            checked = True
        assert checked


def test_create_flat_names_0d():
    shape = ()
    result = ttab.create_flat_names("x", shape)
    expected = ["x"]
    assert result == expected
    assert ttab._create_shape(result) == shape


def test_create_flat_names_1d():
    shape = (2,)
    result = ttab.create_flat_names("x", shape)
    expected = ["x__0", "x__1"]
    assert result == expected
    assert ttab._create_shape(result) == shape


def test_create_flat_names_2d():
    shape = 2, 3
    result = ttab.create_flat_names("x", shape)
    expected = ["x__0_0", "x__0_1", "x__0_2", "x__1_0", "x__1_1", "x__1_2"]
    assert result == expected
    assert ttab._create_shape(result) == shape


def test_create_flat_names_3d():
    shape = 2, 3, 4
    assert ttab._create_shape(ttab.create_flat_names("x", shape)) == shape



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import numpy.testing as npt
import pytest
import theano
import theano.tensor as tt

from numpy import random as nr

import pymc3 as pm

from pymc3.distributions.distribution import _draw_value, draw_values
from pymc3.tests.helpers import SeededTest


def test_draw_value():
    npt.assert_equal(_draw_value(np.array([5, 6])), [5, 6])
    npt.assert_equal(_draw_value(np.array(5.0)), 5)

    npt.assert_equal(_draw_value(tt.constant([5.0, 6.0])), [5, 6])
    assert _draw_value(tt.constant(5)) == 5
    npt.assert_equal(_draw_value(2 * tt.constant([5.0, 6.0])), [10, 12])

    val = theano.shared(np.array([5.0, 6.0]))
    npt.assert_equal(_draw_value(val), [5, 6])
    npt.assert_equal(_draw_value(2 * val), [10, 12])

    a = tt.scalar("a")
    a.tag.test_value = 6
    npt.assert_equal(_draw_value(2 * a, givens=[(a, 1)]), 2)

    assert _draw_value(5) == 5
    assert _draw_value(5.0) == 5
    assert isinstance(_draw_value(5.0), type(5.0))
    assert isinstance(_draw_value(5), type(5))

    with pm.Model():
        mu = 2 * tt.constant(np.array([5.0, 6.0])) + theano.shared(np.array(5))
        a = pm.Normal("a", mu=mu, sigma=5, shape=2)

    val1 = _draw_value(a)
    val2 = _draw_value(a)
    assert np.all(val1 != val2)

    with pytest.raises(ValueError) as err:
        _draw_value([])
    err.match("Unexpected type")


class TestDrawValues:
    def test_empty(self):
        assert draw_values([]) == []

    def test_vals(self):
        npt.assert_equal(draw_values([np.array([5, 6])])[0], [5, 6])
        npt.assert_equal(draw_values([np.array(5.0)])[0], 5)

        npt.assert_equal(draw_values([tt.constant([5.0, 6.0])])[0], [5, 6])
        assert draw_values([tt.constant(5)])[0] == 5
        npt.assert_equal(draw_values([2 * tt.constant([5.0, 6.0])])[0], [10, 12])

        val = theano.shared(np.array([5.0, 6.0]))
        npt.assert_equal(draw_values([val])[0], [5, 6])
        npt.assert_equal(draw_values([2 * val])[0], [10, 12])

    def test_simple_model(self):
        with pm.Model():
            mu = 2 * tt.constant(np.array([5.0, 6.0])) + theano.shared(np.array(5))
            a = pm.Normal("a", mu=mu, sigma=5, shape=2)

        val1 = draw_values([a])
        val2 = draw_values([a])
        assert np.all(val1[0] != val2[0])

        point = {"a": np.array([3.0, 4.0])}
        npt.assert_equal(draw_values([a], point=point), [point["a"]])

    def test_dep_vars(self):
        with pm.Model():
            mu = 2 * tt.constant(np.array([5.0, 6.0])) + theano.shared(np.array(5))
            sd = pm.HalfNormal("sd", shape=2)
            tau = 1 / sd ** 2
            a = pm.Normal("a", mu=mu, tau=tau, shape=2)

        point = {"a": np.array([1.0, 2.0])}
        npt.assert_equal(draw_values([a], point=point), [point["a"]])

        val1 = draw_values([a])[0]
        val2 = draw_values([a], point={"sd": np.array([2.0, 3.0])})[0]
        val3 = draw_values([a], point={"sd_log__": np.array([2.0, 3.0])})[0]
        val4 = draw_values([a], point={"sd_log__": np.array([2.0, 3.0])})[0]

        assert all(
            [
                np.all(val1 != val2),
                np.all(val1 != val3),
                np.all(val1 != val4),
                np.all(val2 != val3),
                np.all(val2 != val4),
                np.all(val3 != val4),
            ]
        )

    def test_graph_constant(self):
        # Issue 3595 pointed out that slice(None) can introduce
        # theano.graph.basic.Constant into the compute graph, which wasn't
        # handled correctly by draw_values
        n_d = 500
        n_x = 2
        n_y = 1
        n_g = 10
        g = np.random.randint(0, n_g, (n_d,))  # group
        x = np.random.randint(0, n_x, (n_d,))  # x factor
        with pm.Model():
            multi_dim_rv = pm.Normal("multi_dim_rv", mu=0, sd=1, shape=(n_x, n_g, n_y))
            indexed_rv = multi_dim_rv[x, g, :]
            i = draw_values([indexed_rv])
            assert i is not None


class TestJointDistributionDrawValues(SeededTest):
    def test_joint_distribution(self):
        with pm.Model() as model:
            a = pm.Normal("a", mu=0, sigma=100)
            b = pm.Normal("b", mu=a, sigma=1e-8)
            c = pm.Normal("c", mu=a, sigma=1e-8)
            d = pm.Deterministic("d", b + c)

        # Expected RVs
        N = 1000
        norm = np.random.randn(3, N)
        eA = norm[0] * 100
        eB = eA + norm[1] * 1e-8
        eC = eA + norm[2] * 1e-8
        eD = eB + eC

        # Drawn RVs
        nr.seed(self.random_seed)
        #        A, B, C, D = list(zip(*[draw_values([a, b, c, d]) for i in range(N)]))
        A, B, C, D = draw_values([a, b, c, d], size=N)
        A = np.array(A).flatten()
        B = np.array(B).flatten()
        C = np.array(C).flatten()
        D = np.array(D).flatten()

        # Assert that the drawn samples match the expected values
        assert np.allclose(eA, A)
        assert np.allclose(eB, B)
        assert np.allclose(eC, C)
        assert np.allclose(eD, D)

        # Assert that A, B and C have the expected difference
        assert np.all(np.abs(A - B) < 1e-6)
        assert np.all(np.abs(A - C) < 1e-6)
        assert np.all(np.abs(B - C) < 1e-6)

        # Marginal draws
        mA = np.array([draw_values([a]) for i in range(N)]).flatten()
        mB = np.array([draw_values([b]) for i in range(N)]).flatten()
        mC = np.array([draw_values([c]) for i in range(N)]).flatten()
        # Also test the with model context of draw_values
        with model:
            mD = np.array([draw_values([d]) for i in range(N)]).flatten()

        # Assert that the marginal distributions have different sample values
        assert not np.all(np.abs(B - mB) < 1e-2)
        assert not np.all(np.abs(C - mC) < 1e-2)
        assert not np.all(np.abs(D - mD) < 1e-2)

        # Assert that the marginal distributions do not have high cross
        # correlation
        assert np.abs(np.corrcoef(mA, mB)[0, 1]) < 0.1
        assert np.abs(np.corrcoef(mA, mC)[0, 1]) < 0.1
        assert np.abs(np.corrcoef(mB, mC)[0, 1]) < 0.1



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import pandas as pd
import pytest
import theano.tensor as tt

from pymc3.glm import utils


class TestUtils:
    def setup_method(self):
        self.data = pd.DataFrame(dict(a=[1, 2, 3], b=[4, 5, 6]))

    def assertMatrixLabels(self, m, l, mt=None, lt=None):
        assert np.all(np.equal(m.eval(), mt if mt is not None else self.data.values))
        assert l == list(lt or self.data.columns)

    def test_numpy_init(self):
        m, l = utils.any_to_tensor_and_labels(self.data.values)
        self.assertMatrixLabels(m, l, lt=["x0", "x1"])
        m, l = utils.any_to_tensor_and_labels(self.data.values, labels=["x2", "x3"])
        self.assertMatrixLabels(m, l, lt=["x2", "x3"])

    def test_pandas_init(self):
        m, l = utils.any_to_tensor_and_labels(self.data)
        self.assertMatrixLabels(m, l)
        m, l = utils.any_to_tensor_and_labels(self.data, labels=["x2", "x3"])
        self.assertMatrixLabels(m, l, lt=["x2", "x3"])

    @pytest.mark.xfail
    def test_dict_input(self):
        m, l = utils.any_to_tensor_and_labels(self.data.to_dict("dict"))
        self.assertMatrixLabels(m, l, mt=self.data[l].values, lt=l)

        m, l = utils.any_to_tensor_and_labels(self.data.to_dict("series"))
        self.assertMatrixLabels(m, l, mt=self.data[l].values, lt=l)

        m, l = utils.any_to_tensor_and_labels(self.data.to_dict("list"))
        self.assertMatrixLabels(m, l, mt=self.data[l].values, lt=l)

        inp = {k: tt.as_tensor_variable(v.values) for k, v in self.data.to_dict("series").items()}
        m, l = utils.any_to_tensor_and_labels(inp)
        self.assertMatrixLabels(m, l, mt=self.data[l].values, lt=l)

    def test_list_input(self):
        m, l = utils.any_to_tensor_and_labels(self.data.values.tolist())
        self.assertMatrixLabels(m, l, lt=["x0", "x1"])
        m, l = utils.any_to_tensor_and_labels(self.data.values.tolist(), labels=["x2", "x3"])
        self.assertMatrixLabels(m, l, lt=["x2", "x3"])

    def test_tensor_input(self):
        m, l = utils.any_to_tensor_and_labels(
            tt.as_tensor_variable(self.data.values.tolist()), labels=["x0", "x1"]
        )
        self.assertMatrixLabels(m, l, lt=["x0", "x1"])
        m, l = utils.any_to_tensor_and_labels(
            tt.as_tensor_variable(self.data.values.tolist()), labels=["x2", "x3"]
        )
        self.assertMatrixLabels(m, l, lt=["x2", "x3"])

    def test_user_mistakes(self):
        # no labels for tensor variable
        with pytest.raises(ValueError):
            utils.any_to_tensor_and_labels(tt.as_tensor_variable(self.data.values.tolist()))
        # len of labels is bad
        with pytest.raises(ValueError):
            utils.any_to_tensor_and_labels(self.data.values.tolist(), labels=["x"])



print("import: 'pymc3'")
import pymc3

print("import: 'pymc3.backends'")
import pymc3.backends

print("import: 'pymc3.distributions'")
import pymc3.distributions

print("import: 'pymc3.glm'")
import pymc3.glm

print("import: 'pymc3.gp'")
import pymc3.gp

print("import: 'pymc3.ode'")
import pymc3.ode

print("import: 'pymc3.plots'")
import pymc3.plots

print("import: 'pymc3.smc'")
import pymc3.smc

print("import: 'pymc3.stats'")
import pymc3.stats

print("import: 'pymc3.step_methods'")
import pymc3.step_methods

print("import: 'pymc3.tests'")
import pymc3.tests

print("import: 'pymc3.tuning'")
import pymc3.tuning

print("import: 'pymc3.variational'")
import pymc3.variational




#   Copyright 2021 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
import numpy as np
import numpy.testing as npt
import pytest
import theano
import theano.tensor as tt

from scipy import interpolate, stats

import pymc3 as pm

from pymc3.distributions import Discrete
from pymc3.distributions.dist_math import (
    MvNormalLogp,
    SplineWrapper,
    alltrue_scalar,
    bound,
    clipped_beta_rvs,
    factln,
    i0e,
)
from pymc3.tests.helpers import verify_grad
from pymc3.theanof import floatX


def test_bound():
    logp = tt.ones((10, 10))
    cond = tt.ones((10, 10))
    assert np.all(bound(logp, cond).eval() == logp.eval())

    logp = tt.ones((10, 10))
    cond = tt.zeros((10, 10))
    assert np.all(bound(logp, cond).eval() == (-np.inf * logp).eval())

    logp = tt.ones((10, 10))
    cond = True
    assert np.all(bound(logp, cond).eval() == logp.eval())

    logp = tt.ones(3)
    cond = np.array([1, 0, 1])
    assert not np.all(bound(logp, cond).eval() == 1)
    assert np.prod(bound(logp, cond).eval()) == -np.inf

    logp = tt.ones((2, 3))
    cond = np.array([[1, 1, 1], [1, 0, 1]])
    assert not np.all(bound(logp, cond).eval() == 1)
    assert np.prod(bound(logp, cond).eval()) == -np.inf


def test_check_bounds_false():
    with pm.Model(check_bounds=False):
        logp = tt.ones(3)
        cond = np.array([1, 0, 1])
        assert np.all(bound(logp, cond).eval() == logp.eval())


def test_alltrue_scalar():
    assert alltrue_scalar([]).eval()
    assert alltrue_scalar([True]).eval()
    assert alltrue_scalar([tt.ones(10)]).eval()
    assert alltrue_scalar([tt.ones(10), 5 * tt.ones(101)]).eval()
    assert alltrue_scalar([np.ones(10), 5 * tt.ones(101)]).eval()
    assert alltrue_scalar([np.ones(10), True, 5 * tt.ones(101)]).eval()
    assert alltrue_scalar([np.array([1, 2, 3]), True, 5 * tt.ones(101)]).eval()

    assert not alltrue_scalar([False]).eval()
    assert not alltrue_scalar([tt.zeros(10)]).eval()
    assert not alltrue_scalar([True, False]).eval()
    assert not alltrue_scalar([np.array([0, -1]), tt.ones(60)]).eval()
    assert not alltrue_scalar([np.ones(10), False, 5 * tt.ones(101)]).eval()


def test_alltrue_shape():
    vals = [True, tt.ones(10), tt.zeros(5)]

    assert alltrue_scalar(vals).eval().shape == ()


class MultinomialA(Discrete):
    def __init__(self, n, p, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.n = n
        self.p = p

    def logp(self, value):
        n = self.n
        p = self.p

        return bound(
            factln(n) - factln(value).sum() + (value * tt.log(p)).sum(),
            value >= 0,
            0 <= p,
            p <= 1,
            tt.isclose(p.sum(), 1),
            broadcast_conditions=False,
        )


class MultinomialB(Discrete):
    def __init__(self, n, p, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.n = n
        self.p = p

    def logp(self, value):
        n = self.n
        p = self.p

        return bound(
            factln(n) - factln(value).sum() + (value * tt.log(p)).sum(),
            tt.all(value >= 0),
            tt.all(0 <= p),
            tt.all(p <= 1),
            tt.isclose(p.sum(), 1),
            broadcast_conditions=False,
        )


def test_multinomial_bound():

    x = np.array([1, 5])
    n = x.sum()

    with pm.Model() as modelA:
        p_a = pm.Dirichlet("p", floatX(np.ones(2)), shape=(2,))
        MultinomialA("x", n, p_a, observed=x)

    with pm.Model() as modelB:
        p_b = pm.Dirichlet("p", floatX(np.ones(2)), shape=(2,))
        MultinomialB("x", n, p_b, observed=x)

    assert np.isclose(
        modelA.logp({"p_stickbreaking__": [0]}), modelB.logp({"p_stickbreaking__": [0]})
    )


class TestMvNormalLogp:
    def test_logp(self):
        np.random.seed(42)

        chol_val = floatX(np.array([[1, 0.9], [0, 2]]))
        cov_val = floatX(np.dot(chol_val, chol_val.T))
        cov = tt.matrix("cov")
        cov.tag.test_value = cov_val
        delta_val = floatX(np.random.randn(5, 2))
        delta = tt.matrix("delta")
        delta.tag.test_value = delta_val
        expect = stats.multivariate_normal(mean=np.zeros(2), cov=cov_val)
        expect = expect.logpdf(delta_val).sum()
        logp = MvNormalLogp()(cov, delta)
        logp_f = theano.function([cov, delta], logp)
        logp = logp_f(cov_val, delta_val)
        npt.assert_allclose(logp, expect)

    @theano.config.change_flags(compute_test_value="ignore")
    def test_grad(self):
        np.random.seed(42)

        def func(chol_vec, delta):
            chol = tt.stack(
                [
                    tt.stack([tt.exp(0.1 * chol_vec[0]), 0]),
                    tt.stack([chol_vec[1], 2 * tt.exp(chol_vec[2])]),
                ]
            )
            cov = tt.dot(chol, chol.T)
            return MvNormalLogp()(cov, delta)

        chol_vec_val = floatX(np.array([0.5, 1.0, -0.1]))

        delta_val = floatX(np.random.randn(1, 2))
        verify_grad(func, [chol_vec_val, delta_val])

        delta_val = floatX(np.random.randn(5, 2))
        verify_grad(func, [chol_vec_val, delta_val])

    @pytest.mark.skip(reason="Fix in theano not released yet: Theano#5908")
    @theano.config.change_flags(compute_test_value="ignore")
    def test_hessian(self):
        chol_vec = tt.vector("chol_vec")
        chol_vec.tag.test_value = np.array([0.1, 2, 3])
        chol = tt.stack(
            [
                tt.stack([tt.exp(0.1 * chol_vec[0]), 0]),
                tt.stack([chol_vec[1], 2 * tt.exp(chol_vec[2])]),
            ]
        )
        cov = tt.dot(chol, chol.T)
        delta = tt.matrix("delta")
        delta.tag.test_value = np.ones((5, 2))
        logp = MvNormalLogp()(cov, delta)
        g_cov, g_delta = tt.grad(logp, [cov, delta])
        tt.grad(g_delta.sum() + g_cov.sum(), [delta, cov])


class TestSplineWrapper:
    @theano.config.change_flags(compute_test_value="ignore")
    def test_grad(self):
        x = np.linspace(0, 1, 100)
        y = x * x
        spline = SplineWrapper(interpolate.InterpolatedUnivariateSpline(x, y, k=1))
        verify_grad(spline, [0.5])

    @theano.config.change_flags(compute_test_value="ignore")
    def test_hessian(self):
        x = np.linspace(0, 1, 100)
        y = x * x
        spline = SplineWrapper(interpolate.InterpolatedUnivariateSpline(x, y, k=1))
        x_var = tt.dscalar("x")
        (g_x,) = tt.grad(spline(x_var), [x_var])
        with pytest.raises(NotImplementedError):
            tt.grad(g_x, [x_var])


class TestI0e:
    @theano.config.change_flags(compute_test_value="ignore")
    def test_grad(self):
        verify_grad(i0e, [0.5])
        verify_grad(i0e, [-2.0])
        verify_grad(i0e, [[0.5, -2.0]])
        verify_grad(i0e, [[[0.5, -2.0]]])


@pytest.mark.parametrize("dtype", ["float16", "float32", "float64"])
def test_clipped_beta_rvs(dtype):
    # Verify that the samples drawn from the beta distribution are never
    # equal to zero or one (issue #3898)
    values = clipped_beta_rvs(0.01, 0.01, size=1000000, dtype=dtype)
    assert not (np.any(values == 0) or np.any(values == 1))



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import pytest
import theano

from cachetools import cached
from numpy.testing import assert_almost_equal

import pymc3 as pm

from pymc3.distributions.transforms import Transform
from pymc3.tests.helpers import SeededTest
from pymc3.util import hash_key, hashable, locally_cachedmethod


class TestBackendVersionCheck:
    def test_warn_on_incompatible_backend(self, capsys):
        assert not "!!!!!" in capsys.readouterr().out
        pm._check_backend_version()
        assert not "!!!!!" in capsys.readouterr().out

        # Mock an incorrect backend version
        original = theano.__version__

        theano.__version__ = "1.1.0"
        pm._check_backend_version()
        assert "does not match" in capsys.readouterr().out

        del theano.__version__
        pm._check_backend_version()
        assert "is broken" in capsys.readouterr().out

        theano.__version__ = original
        pass


class TestTransformName:
    cases = [("var", "var_test__"), ("var_test_", "var_test__test__")]
    transform_name = "test"

    def test_get_transformed_name(self):
        test_transform = Transform()
        test_transform.name = self.transform_name
        for name, transformed in self.cases:
            assert pm.util.get_transformed_name(name, test_transform) == transformed

    def test_is_transformed_name(self):
        for name, transformed in self.cases:
            assert pm.util.is_transformed_name(transformed)
            assert not pm.util.is_transformed_name(name)

    def test_get_untransformed_name(self):
        for name, transformed in self.cases:
            assert pm.util.get_untransformed_name(transformed) == name
            with pytest.raises(ValueError):
                pm.util.get_untransformed_name(name)


class TestUpdateStartVals(SeededTest):
    def setup_method(self):
        super().setup_method()

    def test_soft_update_all_present(self):
        start = {"a": 1, "b": 2}
        test_point = {"a": 3, "b": 4}
        pm.util.update_start_vals(start, test_point, model=None)
        assert start == {"a": 1, "b": 2}

    def test_soft_update_one_missing(self):
        start = {
            "a": 1,
        }
        test_point = {"a": 3, "b": 4}
        pm.util.update_start_vals(start, test_point, model=None)
        assert start == {"a": 1, "b": 4}

    def test_soft_update_empty(self):
        start = {}
        test_point = {"a": 3, "b": 4}
        pm.util.update_start_vals(start, test_point, model=None)
        assert start == test_point

    def test_soft_update_transformed(self):
        with pm.Model() as model:
            pm.Exponential("a", 1)
        start = {"a": 2.0}
        test_point = {"a_log__": 0}
        pm.util.update_start_vals(start, test_point, model)
        assert_almost_equal(np.exp(start["a_log__"]), start["a"])

    def test_soft_update_parent(self):
        with pm.Model() as model:
            a = pm.Uniform("a", lower=0.0, upper=1.0)
            b = pm.Uniform("b", lower=2.0, upper=3.0)
            pm.Uniform("lower", lower=a, upper=3.0)
            pm.Uniform("upper", lower=0.0, upper=b)
            pm.Uniform("interv", lower=a, upper=b)

        start = {"a": 0.3, "b": 2.1, "lower": 1.4, "upper": 1.4, "interv": 1.4}
        test_point = {
            "lower_interval__": -0.3746934494414109,
            "upper_interval__": 0.693147180559945,
            "interv_interval__": 0.4519851237430569,
        }
        pm.util.update_start_vals(start, model.test_point, model)
        assert_almost_equal(start["lower_interval__"], test_point["lower_interval__"])
        assert_almost_equal(start["upper_interval__"], test_point["upper_interval__"])
        assert_almost_equal(start["interv_interval__"], test_point["interv_interval__"])


class TestCheckStartVals(SeededTest):
    def setup_method(self):
        super().setup_method()

    def test_valid_start_point(self):
        with pm.Model() as model:
            a = pm.Uniform("a", lower=0.0, upper=1.0)
            b = pm.Uniform("b", lower=2.0, upper=3.0)

        start = {"a": 0.3, "b": 2.1}
        pm.util.update_start_vals(start, model.test_point, model)
        pm.util.check_start_vals(start, model)

    def test_invalid_start_point(self):
        with pm.Model() as model:
            a = pm.Uniform("a", lower=0.0, upper=1.0)
            b = pm.Uniform("b", lower=2.0, upper=3.0)

        start = {"a": np.nan, "b": np.nan}
        pm.util.update_start_vals(start, model.test_point, model)
        with pytest.raises(pm.exceptions.SamplingError):
            pm.util.check_start_vals(start, model)

    def test_invalid_variable_name(self):
        with pm.Model() as model:
            a = pm.Uniform("a", lower=0.0, upper=1.0)
            b = pm.Uniform("b", lower=2.0, upper=3.0)

        start = {"a": 0.3, "b": 2.1, "c": 1.0}
        pm.util.update_start_vals(start, model.test_point, model)
        with pytest.raises(KeyError):
            pm.util.check_start_vals(start, model)


class TestExceptions:
    def test_shape_error(self):
        with pytest.raises(pm.exceptions.ShapeError) as exinfo:
            raise pm.exceptions.ShapeError("Just the message.")
        assert "Just" in exinfo.value.args[0]

        with pytest.raises(pm.exceptions.ShapeError) as exinfo:
            raise pm.exceptions.ShapeError("With shapes.", actual=(2, 3))
        assert "(2, 3)" in exinfo.value.args[0]

        with pytest.raises(pm.exceptions.ShapeError) as exinfo:
            raise pm.exceptions.ShapeError("With shapes.", expected="(2,3) or (5,6)")
        assert "(5,6)" in exinfo.value.args[0]

        with pytest.raises(pm.exceptions.ShapeError) as exinfo:
            raise pm.exceptions.ShapeError("With shapes.", actual=(), expected="(5,4) or (?,?,6)")
        assert "(?,?,6)" in exinfo.value.args[0]
        pass

    def test_dtype_error(self):
        with pytest.raises(pm.exceptions.DtypeError) as exinfo:
            raise pm.exceptions.DtypeError("Just the message.")
        assert "Just" in exinfo.value.args[0]

        with pytest.raises(pm.exceptions.DtypeError) as exinfo:
            raise pm.exceptions.DtypeError("With types.", actual=str)
        assert "str" in exinfo.value.args[0]

        with pytest.raises(pm.exceptions.DtypeError) as exinfo:
            raise pm.exceptions.DtypeError("With types.", expected=float)
        assert "float" in exinfo.value.args[0]

        with pytest.raises(pm.exceptions.DtypeError) as exinfo:
            raise pm.exceptions.DtypeError("With types.", actual=int, expected=str)
        assert "int" in exinfo.value.args[0] and "str" in exinfo.value.args[0]
        pass


def test_hashing_of_rv_tuples():
    obs = np.random.normal(-1, 0.1, size=10)
    with pm.Model() as pmodel:
        mu = pm.Normal("mu", 0, 1)
        sd = pm.Gamma("sd", 1, 2)
        dd = pm.DensityDist(
            "dd",
            pm.Normal.dist(mu, sd).logp,
            random=pm.Normal.dist(mu, sd).random,
            observed=obs,
        )
        for freerv in [mu, sd, dd] + pmodel.free_RVs:
            for structure in [
                freerv,
                {"alpha": freerv, "omega": None},
                [freerv, []],
                (freerv, []),
            ]:
                assert isinstance(hashable(structure), int)


def test_hash_key():
    class Bad1:
        def __hash__(self):
            return 329

    class Bad2:
        def __hash__(self):
            return 329

    b1 = Bad1()
    b2 = Bad2()

    assert b1 != b2

    @cached({}, key=hash_key)
    def some_func(x):
        return x

    assert some_func(b1) != some_func(b2)

    class TestClass:
        @locally_cachedmethod
        def some_method(self, x):
            return x

    tc = TestClass()
    assert tc.some_method(b1) != tc.some_method(b2)



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import shutil
import sys
import tempfile

from math import isclose

import arviz as az
import numpy as np
import numpy.testing as npt
import pytest
import theano
import theano.tensor as tt

from numpy.testing import assert_array_almost_equal
from theano.compile.ops import as_op
from theano.graph.op import Op

from pymc3.data import Data
from pymc3.distributions import (
    Bernoulli,
    Beta,
    Binomial,
    Categorical,
    HalfNormal,
    MvNormal,
    Normal,
)
from pymc3.exceptions import SamplingError
from pymc3.model import Model, Potential, set_data
from pymc3.sampling import assign_step_methods, sample
from pymc3.step_methods import (
    MLDA,
    NUTS,
    BinaryGibbsMetropolis,
    CategoricalGibbsMetropolis,
    CompoundStep,
    DEMetropolis,
    DEMetropolisZ,
    EllipticalSlice,
    HamiltonianMC,
    Metropolis,
    MultivariateNormalProposal,
    NormalProposal,
    RecursiveDAProposal,
    Slice,
    UniformProposal,
)
from pymc3.step_methods.mlda import extract_Q_estimate
from pymc3.tests.checks import close_to
from pymc3.tests.helpers import select_by_precision
from pymc3.tests.models import (
    mv_prior_simple,
    mv_simple,
    mv_simple_coarse,
    mv_simple_discrete,
    mv_simple_very_coarse,
    simple_2model_continuous,
    simple_categorical,
)
from pymc3.theanof import floatX


class TestStepMethods:  # yield test doesn't work subclassing object
    master_samples = {
        Slice: np.array(
            [
                0.10233528,
                0.40458486,
                0.17329217,
                0.46281232,
                0.22556278,
                1.52632836,
                -0.27823807,
                0.02539625,
                1.02711735,
                0.03686346,
                -0.62841281,
                -0.27125083,
                0.31989505,
                0.84031155,
                -0.18949138,
                1.60550262,
                1.01375291,
                -0.29742941,
                0.35312738,
                0.43363622,
                1.18898078,
                0.80063888,
                0.38445644,
                0.90184395,
                1.69150017,
                2.05452171,
                -0.13334755,
                1.61265408,
                1.36579345,
                1.3216292,
                -0.59487037,
                -0.34648927,
                1.05107285,
                0.42870305,
                0.61552257,
                0.55239884,
                0.13929271,
                0.26213809,
                -0.2316028,
                0.19711046,
                1.42832629,
                1.93641434,
                -0.81142379,
                -0.31059485,
                -0.3189694,
                1.43542534,
                0.40311093,
                1.63103768,
                0.24034874,
                0.33924866,
                0.94951616,
                0.71700185,
                0.79273056,
                -0.44569146,
                1.91974783,
                0.84673795,
                1.12411833,
                -0.83123811,
                -0.54310095,
                -0.00721347,
                0.9925055,
                1.04015058,
                -0.34958074,
                -0.14926302,
                -0.47990225,
                -0.75629446,
                -0.95942067,
                1.68179204,
                1.20598073,
                1.39675733,
                1.22755935,
                0.06728757,
                1.05184231,
                1.01126791,
                -0.67327093,
                0.21429651,
                1.33730461,
                -1.56174184,
                -0.64348764,
                0.98050636,
                0.25923049,
                0.58622631,
                0.46589069,
                1.44367347,
                -0.43141573,
                1.08293374,
                -0.5563204,
                1.46287904,
                1.26019815,
                0.52972104,
                1.08792687,
                1.10064358,
                1.84881549,
                0.91179647,
                0.69316592,
                -0.47657064,
                2.22747063,
                0.83388935,
                0.84680716,
                -0.10556406,
            ]
        ),
        HamiltonianMC: np.array(
            [
                1.43583525,
                1.43583525,
                1.43583525,
                -0.57415005,
                0.91472062,
                0.91472062,
                0.36282799,
                0.80991631,
                0.84457253,
                0.84457253,
                -0.12651784,
                -0.12651784,
                0.39027088,
                -0.22998424,
                0.64337475,
                0.64337475,
                0.03504003,
                1.2667789,
                1.2667789,
                0.34770874,
                0.224319,
                0.224319,
                1.00416894,
                0.46161403,
                0.28217305,
                0.28217305,
                0.50327811,
                0.50327811,
                0.50327811,
                0.50327811,
                0.42335724,
                0.42335724,
                0.20336198,
                0.20336198,
                0.20336198,
                0.16330229,
                0.16330229,
                -0.7332075,
                1.04924226,
                1.04924226,
                0.39630439,
                0.16481719,
                0.16481719,
                0.84146061,
                0.83146709,
                0.83146709,
                0.32748059,
                1.00918804,
                1.00918804,
                0.91034823,
                1.31278027,
                1.38222654,
                1.38222654,
                -0.32268814,
                -0.32268814,
                2.1866116,
                1.21679252,
                -0.15916878,
                -0.15916878,
                0.38958249,
                0.38958249,
                0.54971928,
                0.05591406,
                0.87712017,
                0.87712017,
                0.19409043,
                0.19409043,
                0.19409043,
                0.40718849,
                0.63399349,
                0.35510353,
                0.35510353,
                0.47860847,
                0.47860847,
                0.69805772,
                0.16686305,
                0.16686305,
                0.16686305,
                0.04971251,
                0.04971251,
                -0.90052793,
                -0.73203754,
                1.02258958,
                1.02258958,
                -0.14144856,
                -0.14144856,
                1.43017486,
                1.23202605,
                1.23202605,
                0.24442885,
                0.78300516,
                0.30494261,
                0.30494261,
                0.30494261,
                -0.00596443,
                1.31695235,
                0.81375848,
                0.81375848,
                0.81375848,
                1.91238675,
            ]
        ),
        Metropolis: np.array(
            [
                1.62434536,
                1.01258895,
                0.4844172,
                0.4844172,
                0.4844172,
                0.4844172,
                0.4844172,
                0.4844172,
                0.4844172,
                0.4844172,
                0.31198899,
                0.31198899,
                0.31198899,
                0.31198899,
                1.21284494,
                0.52911708,
                0.261229,
                0.79158447,
                0.10441177,
                -0.74079387,
                -0.74079387,
                -0.50637818,
                -0.50637818,
                -0.50637818,
                -0.45557042,
                -0.45557042,
                -0.33541147,
                0.28179164,
                0.58196196,
                0.22971211,
                0.02081788,
                0.60744107,
                0.8930284,
                0.8930284,
                1.40595822,
                1.10786538,
                1.10786538,
                1.10786538,
                1.10786538,
                -0.28863095,
                -0.12859388,
                0.74757504,
                0.74757504,
                0.74757504,
                0.97766977,
                0.97766977,
                0.75534163,
                0.55458356,
                0.75288328,
                0.87189193,
                0.9937132,
                0.9937132,
                0.61842825,
                0.61842825,
                0.27457457,
                0.31817143,
                0.31817143,
                0.31817143,
                -0.77674042,
                -0.60735798,
                0.13319847,
                -0.82050213,
                -0.82050213,
                -0.50534274,
                -0.15479676,
                -0.15479676,
                -0.19349227,
                -0.19349227,
                -0.21810923,
                -0.21810923,
                -0.21810923,
                1.0180548,
                -0.18121323,
                0.68213209,
                0.68213209,
                1.23266958,
                1.23266958,
                0.60913885,
                1.41099989,
                1.45756718,
                1.45756718,
                1.45756718,
                1.45756718,
                1.59526839,
                1.82776295,
                1.82776295,
                1.82776295,
                1.82776295,
                2.2691274,
                2.16897216,
                2.18638157,
                1.06436284,
                0.54726838,
                0.54726838,
                1.04247971,
                0.86777655,
                0.86777655,
                0.86777655,
                0.86777655,
                0.61914177,
            ]
        ),
        NUTS: np.array(
            [
                0.550575,
                0.550575,
                0.80031201,
                0.91580544,
                1.34622953,
                1.34622953,
                -0.63861533,
                -0.62101385,
                -0.62101385,
                -0.60250375,
                -1.04753424,
                -0.34850626,
                0.35882649,
                -0.20339408,
                -0.18077466,
                -0.18077466,
                0.1242007,
                -0.48708213,
                0.01216292,
                0.01216292,
                -0.15991487,
                0.0118306,
                0.0118306,
                0.02512962,
                -0.06002705,
                0.61278464,
                -0.45991609,
                -0.45991609,
                -0.45991609,
                -0.3067988,
                -0.3067988,
                -0.30830273,
                -0.62877494,
                -0.5896293,
                0.32740518,
                0.32740518,
                0.55321326,
                0.34885231,
                0.34885231,
                0.35304997,
                1.20016133,
                1.20016133,
                1.26432486,
                1.22481613,
                1.46040499,
                1.2251786,
                0.29954482,
                0.29954482,
                0.5713582,
                0.5755183,
                0.26968846,
                0.68253483,
                0.68253483,
                0.69418724,
                1.4172782,
                1.4172782,
                0.85063608,
                0.23409974,
                -0.65012501,
                1.16211157,
                -0.04844954,
                1.34390994,
                -0.44058335,
                -0.44058335,
                0.85096033,
                0.98734074,
                1.31200906,
                1.2751574,
                1.2751574,
                0.04377635,
                0.08244824,
                0.6342471,
                -0.31243596,
                1.0165907,
                -0.19025897,
                -0.19025897,
                0.02133041,
                -0.02335463,
                0.43923434,
                -0.45033488,
                0.05985518,
                -0.10019701,
                1.34229104,
                1.28571862,
                0.59557205,
                0.63730268,
                0.63730268,
                0.54269992,
                0.54269992,
                -0.48334519,
                1.02199273,
                -0.17367903,
                -0.17367903,
                0.8470911,
                -0.12868214,
                1.8986946,
                1.55412619,
                1.55412619,
                0.90228003,
                1.3328478,
            ]
        ),
    }

    def setup_class(self):
        self.temp_dir = tempfile.mkdtemp()

    def teardown_class(self):
        shutil.rmtree(self.temp_dir)

    @pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")
    def test_sample_exact(self):
        for step_method in self.master_samples:
            self.check_trace(step_method)

    def check_trace(self, step_method):
        """Tests whether the trace for step methods is exactly the same as on master.

        Code changes that effect how random numbers are drawn may change this, and require
        `master_samples` to be updated, but such changes should be noted and justified in the
        commit.

        This method may also be used to benchmark step methods across commits, by running, for
        example

        ```
        BENCHMARK=100000 ./scripts/test.sh -s pymc3/tests/test_step.py:TestStepMethods
        ```

        on multiple commits.
        """
        n_steps = 100
        with Model() as model:
            x = Normal("x", mu=0, sigma=1)
            y = Normal("y", mu=x, sigma=1, observed=1)
            if step_method.__name__ == "NUTS":
                step = step_method(scaling=model.test_point)
                trace = sample(
                    0, tune=n_steps, discard_tuned_samples=False, step=step, random_seed=1, chains=1
                )
            else:
                trace = sample(
                    0,
                    tune=n_steps,
                    discard_tuned_samples=False,
                    step=step_method(),
                    random_seed=1,
                    chains=1,
                )

        assert_array_almost_equal(
            trace["x"],
            self.master_samples[step_method],
            decimal=select_by_precision(float64=6, float32=4),
        )

    def check_stat(self, check, trace, name):
        for (var, stat, value, bound) in check:
            s = stat(trace[var][2000:], axis=0)
            close_to(s, value, bound)

    def test_step_continuous(self):
        start, model, (mu, C) = mv_simple()
        unc = np.diag(C) ** 0.5
        check = (("x", np.mean, mu, unc / 10.0), ("x", np.std, unc, unc / 10.0))
        _, model_coarse, _ = mv_simple_coarse()
        with model:
            steps = (
                Slice(),
                HamiltonianMC(scaling=C, is_cov=True, blocked=False),
                NUTS(scaling=C, is_cov=True, blocked=False),
                Metropolis(S=C, proposal_dist=MultivariateNormalProposal, blocked=True),
                Slice(blocked=True),
                HamiltonianMC(scaling=C, is_cov=True),
                NUTS(scaling=C, is_cov=True),
                CompoundStep(
                    [
                        HamiltonianMC(scaling=C, is_cov=True),
                        HamiltonianMC(scaling=C, is_cov=True, blocked=False),
                    ]
                ),
                MLDA(
                    coarse_models=[model_coarse],
                    base_S=C,
                    base_proposal_dist=MultivariateNormalProposal,
                ),
            )
        for step in steps:
            trace = sample(
                0,
                tune=8000,
                chains=1,
                discard_tuned_samples=False,
                step=step,
                start=start,
                model=model,
                random_seed=1,
            )
            self.check_stat(check, trace, step.__class__.__name__)

    def test_step_discrete(self):
        if theano.config.floatX == "float32":
            return  # Cannot use @skip because it only skips one iteration of the yield
        start, model, (mu, C) = mv_simple_discrete()
        unc = np.diag(C) ** 0.5
        check = (("x", np.mean, mu, unc / 10.0), ("x", np.std, unc, unc / 10.0))
        with model:
            steps = (Metropolis(S=C, proposal_dist=MultivariateNormalProposal),)
        for step in steps:
            trace = sample(
                20000, tune=0, step=step, start=start, model=model, random_seed=1, chains=1
            )
            self.check_stat(check, trace, step.__class__.__name__)

    def test_step_categorical(self):
        start, model, (mu, C) = simple_categorical()
        unc = C ** 0.5
        check = (("x", np.mean, mu, unc / 10.0), ("x", np.std, unc, unc / 10.0))
        with model:
            steps = (
                CategoricalGibbsMetropolis(model.x, proposal="uniform"),
                CategoricalGibbsMetropolis(model.x, proposal="proportional"),
            )
        for step in steps:
            trace = sample(8000, tune=0, step=step, start=start, model=model, random_seed=1)
            self.check_stat(check, trace, step.__class__.__name__)

    def test_step_elliptical_slice(self):
        start, model, (K, L, mu, std, noise) = mv_prior_simple()
        unc = noise ** 0.5
        check = (("x", np.mean, mu, unc / 10.0), ("x", np.std, std, unc / 10.0))
        with model:
            steps = (EllipticalSlice(prior_cov=K), EllipticalSlice(prior_chol=L))
        for step in steps:
            trace = sample(
                5000, tune=0, step=step, start=start, model=model, random_seed=1, chains=1
            )
            self.check_stat(check, trace, step.__class__.__name__)


class TestMetropolisProposal:
    def test_proposal_choice(self):
        _, model, _ = mv_simple()
        with model:
            s = np.ones(model.ndim)
            sampler = Metropolis(S=s)
            assert isinstance(sampler.proposal_dist, NormalProposal)
            s = np.diag(s)
            sampler = Metropolis(S=s)
            assert isinstance(sampler.proposal_dist, MultivariateNormalProposal)
            s[0, 0] = -s[0, 0]
            with pytest.raises(np.linalg.LinAlgError):
                sampler = Metropolis(S=s)

    def test_mv_proposal(self):
        np.random.seed(42)
        cov = np.random.randn(5, 5)
        cov = cov.dot(cov.T)
        prop = MultivariateNormalProposal(cov)
        samples = np.array([prop() for _ in range(10000)])
        npt.assert_allclose(np.cov(samples.T), cov, rtol=0.2)


class TestCompoundStep:
    samplers = (Metropolis, Slice, HamiltonianMC, NUTS, DEMetropolis)

    @pytest.mark.skipif(
        theano.config.floatX == "float32", reason="Test fails on 32 bit due to linalg issues"
    )
    def test_non_blocked(self):
        """Test that samplers correctly create non-blocked compound steps."""
        _, model = simple_2model_continuous()
        with model:
            for sampler in self.samplers:
                assert isinstance(sampler(blocked=False), CompoundStep)

    @pytest.mark.skipif(
        theano.config.floatX == "float32", reason="Test fails on 32 bit due to linalg issues"
    )
    def test_blocked(self):
        _, model = simple_2model_continuous()
        with model:
            for sampler in self.samplers:
                sampler_instance = sampler(blocked=True)
                assert not isinstance(sampler_instance, CompoundStep)
                assert isinstance(sampler_instance, sampler)


class TestAssignStepMethods:
    def test_bernoulli(self):
        """Test bernoulli distribution is assigned binary gibbs metropolis method"""
        with Model() as model:
            Bernoulli("x", 0.5)
            steps = assign_step_methods(model, [])
        assert isinstance(steps, BinaryGibbsMetropolis)

    def test_normal(self):
        """Test normal distribution is assigned NUTS method"""
        with Model() as model:
            Normal("x", 0, 1)
            steps = assign_step_methods(model, [])
        assert isinstance(steps, NUTS)

    def test_categorical(self):
        """Test categorical distribution is assigned categorical gibbs metropolis method"""
        with Model() as model:
            Categorical("x", np.array([0.25, 0.75]))
            steps = assign_step_methods(model, [])
        assert isinstance(steps, BinaryGibbsMetropolis)
        with Model() as model:
            Categorical("y", np.array([0.25, 0.70, 0.05]))
            steps = assign_step_methods(model, [])
        assert isinstance(steps, CategoricalGibbsMetropolis)

    def test_binomial(self):
        """Test binomial distribution is assigned metropolis method."""
        with Model() as model:
            Binomial("x", 10, 0.5)
            steps = assign_step_methods(model, [])
        assert isinstance(steps, Metropolis)

    def test_normal_nograd_op(self):
        """Test normal distribution without an implemented gradient is assigned slice method"""
        with Model() as model:
            x = Normal("x", 0, 1)

            # a custom Theano Op that does not have a grad:
            is_64 = theano.config.floatX == "float64"
            itypes = [tt.dscalar] if is_64 else [tt.fscalar]
            otypes = [tt.dscalar] if is_64 else [tt.fscalar]

            @as_op(itypes, otypes)
            def kill_grad(x):
                return x

            data = np.random.normal(size=(100,))
            Normal("y", mu=kill_grad(x), sigma=1, observed=data.astype(theano.config.floatX))

            steps = assign_step_methods(model, [])
        assert isinstance(steps, Slice)


class TestPopulationSamplers:

    steppers = [DEMetropolis]

    def test_checks_population_size(self):
        """Test that population samplers check the population size."""
        with Model() as model:
            n = Normal("n", mu=0, sigma=1)
            for stepper in TestPopulationSamplers.steppers:
                step = stepper()
                with pytest.raises(ValueError):
                    sample(draws=10, tune=10, chains=1, cores=1, step=step)
                # don't parallelize to make test faster
                sample(draws=10, tune=10, chains=4, cores=1, step=step)
        pass

    def test_demcmc_warning_on_small_populations(self):
        """Test that a warning is raised when n_chains <= n_dims"""
        with Model() as model:
            Normal("n", mu=0, sigma=1, shape=(2, 3))
            with pytest.warns(UserWarning) as record:
                sample(
                    draws=5,
                    tune=5,
                    chains=6,
                    step=DEMetropolis(),
                    # make tests faster by not parallelizing; disable convergence warning
                    cores=1,
                    compute_convergence_checks=False,
                )
        pass

    def test_demcmc_tune_parameter(self):
        """Tests that validity of the tune setting is checked"""
        with Model() as model:
            Normal("n", mu=0, sigma=1, shape=(2, 3))

            step = DEMetropolis()
            assert step.tune is None

            step = DEMetropolis(tune="scaling")
            assert step.tune == "scaling"

            step = DEMetropolis(tune="lambda")
            assert step.tune == "lambda"

            with pytest.raises(ValueError):
                DEMetropolis(tune="foo")
        pass

    def test_nonparallelized_chains_are_random(self):
        with Model() as model:
            x = Normal("x", 0, 1)
            for stepper in TestPopulationSamplers.steppers:
                step = stepper()
                trace = sample(chains=4, cores=1, draws=20, tune=0, step=DEMetropolis())
                samples = np.array(trace.get_values("x", combine=False))[:, 5]

                assert len(set(samples)) == 4, "Parallelized {} " "chains are identical.".format(
                    stepper
                )
        pass

    def test_parallelized_chains_are_random(self):
        with Model() as model:
            x = Normal("x", 0, 1)
            for stepper in TestPopulationSamplers.steppers:
                step = stepper()
                trace = sample(chains=4, cores=4, draws=20, tune=0, step=DEMetropolis())
                samples = np.array(trace.get_values("x", combine=False))[:, 5]

                assert len(set(samples)) == 4, "Parallelized {} " "chains are identical.".format(
                    stepper
                )
        pass


class TestMetropolis:
    def test_tuning_reset(self):
        """Re-use of the step method instance with cores=1 must not leak tuning information between chains."""
        with Model() as pmodel:
            D = 3
            Normal("n", 0, 2, shape=(D,))
            trace = sample(
                tune=600,
                draws=500,
                step=Metropolis(tune=True, scaling=0.1),
                cores=1,
                chains=3,
                discard_tuned_samples=False,
            )
        for c in range(trace.nchains):
            # check that the tuned settings changed and were reset
            assert trace.get_sampler_stats("scaling", chains=c)[0] == 0.1
            assert trace.get_sampler_stats("scaling", chains=c)[-1] != 0.1
        pass


class TestDEMetropolisZ:
    def test_tuning_lambda_sequential(self):
        with Model() as pmodel:
            Normal("n", 0, 2, shape=(3,))
            trace = sample(
                tune=1000,
                draws=500,
                step=DEMetropolisZ(tune="lambda", lamb=0.92),
                cores=1,
                chains=3,
                discard_tuned_samples=False,
            )
        for c in range(trace.nchains):
            # check that the tuned settings changed and were reset
            assert trace.get_sampler_stats("lambda", chains=c)[0] == 0.92
            assert trace.get_sampler_stats("lambda", chains=c)[-1] != 0.92
            assert set(trace.get_sampler_stats("tune", chains=c)) == {True, False}
        pass

    def test_tuning_epsilon_parallel(self):
        with Model() as pmodel:
            Normal("n", 0, 2, shape=(3,))
            trace = sample(
                tune=1000,
                draws=500,
                step=DEMetropolisZ(tune="scaling", scaling=0.002),
                cores=2,
                chains=2,
                discard_tuned_samples=False,
            )
        for c in range(trace.nchains):
            # check that the tuned settings changed and were reset
            assert trace.get_sampler_stats("scaling", chains=c)[0] == 0.002
            assert trace.get_sampler_stats("scaling", chains=c)[-1] != 0.002
            assert set(trace.get_sampler_stats("tune", chains=c)) == {True, False}
        pass

    def test_tuning_none(self):
        with Model() as pmodel:
            Normal("n", 0, 2, shape=(3,))
            trace = sample(
                tune=1000,
                draws=500,
                step=DEMetropolisZ(tune=None),
                cores=1,
                chains=2,
                discard_tuned_samples=False,
            )
        for c in range(trace.nchains):
            # check that all tunable parameters remained constant
            assert len(set(trace.get_sampler_stats("lambda", chains=c))) == 1
            assert len(set(trace.get_sampler_stats("scaling", chains=c))) == 1
            assert set(trace.get_sampler_stats("tune", chains=c)) == {True, False}
        pass

    def test_tuning_reset(self):
        """Re-use of the step method instance with cores=1 must not leak tuning information between chains."""
        with Model() as pmodel:
            D = 3
            Normal("n", 0, 2, shape=(D,))
            trace = sample(
                tune=1000,
                draws=500,
                step=DEMetropolisZ(tune="scaling", scaling=0.002),
                cores=1,
                chains=3,
                discard_tuned_samples=False,
            )
        for c in range(trace.nchains):
            # check that the tuned settings changed and were reset
            assert trace.get_sampler_stats("scaling", chains=c)[0] == 0.002
            assert trace.get_sampler_stats("scaling", chains=c)[-1] != 0.002
            # check that the variance of the first 50 iterations is much lower than the last 100
            for d in range(D):
                var_start = np.var(trace.get_values("n", chains=c)[:50, d])
                var_end = np.var(trace.get_values("n", chains=c)[-100:, d])
                assert var_start < 0.1 * var_end
        pass

    def test_tune_drop_fraction(self):
        tune = 300
        tune_drop_fraction = 0.85
        draws = 200
        with Model() as pmodel:
            Normal("n", 0, 2, shape=(3,))
            step = DEMetropolisZ(tune_drop_fraction=tune_drop_fraction)
            trace = sample(
                tune=tune, draws=draws, step=step, cores=1, chains=1, discard_tuned_samples=False
            )
            assert len(trace) == tune + draws
            assert len(step._history) == (tune - tune * tune_drop_fraction) + draws
        pass

    @pytest.mark.parametrize(
        "variable,has_grad,outcome",
        [("n", True, 1), ("n", False, 1), ("b", True, 0), ("b", False, 0)],
    )
    def test_competence(self, variable, has_grad, outcome):
        with Model() as pmodel:
            Normal("n", 0, 2, shape=(3,))
            Binomial("b", n=2, p=0.3)
        assert DEMetropolisZ.competence(pmodel[variable], has_grad=has_grad) == outcome
        pass

    @pytest.mark.parametrize("tune_setting", ["foo", True, False])
    def test_invalid_tune(self, tune_setting):
        with Model() as pmodel:
            Normal("n", 0, 2, shape=(3,))
            with pytest.raises(ValueError):
                DEMetropolisZ(tune=tune_setting)
        pass

    def test_custom_proposal_dist(self):
        with Model() as pmodel:
            D = 3
            Normal("n", 0, 2, shape=(D,))
            trace = sample(
                tune=100,
                draws=50,
                step=DEMetropolisZ(proposal_dist=NormalProposal),
                cores=1,
                chains=3,
                discard_tuned_samples=False,
            )
        pass


@pytest.mark.xfail(condition=(theano.config.floatX == "float32"), reason="Fails on float32")
class TestNutsCheckTrace:
    def test_multiple_samplers(self, caplog):
        with Model():
            prob = Beta("prob", alpha=5.0, beta=3.0)
            Binomial("outcome", n=1, p=prob)
            caplog.clear()
            sample(3, tune=2, discard_tuned_samples=False, n_init=None, chains=1)
            messages = [msg.msg for msg in caplog.records]
            assert all("boolean index did not" not in msg for msg in messages)

    def test_bad_init_nonparallel(self):
        with Model():
            HalfNormal("a", sigma=1, testval=-1, transform=None)
            with pytest.raises(SamplingError) as error:
                sample(init=None, chains=1, random_seed=1)
            error.match("Initial evaluation")

    @pytest.mark.skipif(sys.version_info < (3, 6), reason="requires python3.6 or higher")
    def test_bad_init_parallel(self):
        with Model():
            HalfNormal("a", sigma=1, testval=-1, transform=None)
            with pytest.raises(SamplingError) as error:
                sample(init=None, cores=2, random_seed=1)
            error.match("Initial evaluation")

    def test_linalg(self, caplog):
        with Model():
            a = Normal("a", shape=2)
            a = tt.switch(a > 0, np.inf, a)
            b = tt.slinalg.solve(floatX(np.eye(2)), a)
            Normal("c", mu=b, shape=2)
            caplog.clear()
            trace = sample(20, init=None, tune=5, chains=2)
            warns = [msg.msg for msg in caplog.records]
            assert np.any(trace["diverging"])
            assert (
                any("divergence after tuning" in warn for warn in warns)
                or any("divergences after tuning" in warn for warn in warns)
                or any("only diverging samples" in warn for warn in warns)
            )

            with pytest.raises(ValueError) as error:
                trace.report.raise_ok()
            error.match("issues during sampling")

            assert not trace.report.ok

    def test_sampler_stats(self):
        with Model() as model:
            Normal("x", mu=0, sigma=1)
            trace = sample(draws=10, tune=1, chains=1)

        # Assert stats exist and have the correct shape.
        expected_stat_names = {
            "depth",
            "diverging",
            "energy",
            "energy_error",
            "model_logp",
            "max_energy_error",
            "mean_tree_accept",
            "step_size",
            "step_size_bar",
            "tree_size",
            "tune",
            "perf_counter_diff",
            "perf_counter_start",
            "process_time_diff",
        }
        assert trace.stat_names == expected_stat_names
        for varname in trace.stat_names:
            assert trace.get_sampler_stats(varname).shape == (10,)

        # Assert model logp is computed correctly: computing post-sampling
        # and tracking while sampling should give same results.
        model_logp_ = np.array(
            [model.logp(trace.point(i, chain=c)) for c in trace.chains for i in range(len(trace))]
        )
        assert (trace.model_logp == model_logp_).all()


class TestMLDA:
    steppers = [MLDA]

    def test_proposal_and_base_proposal_choice(self):
        """Test that proposal_dist and base_proposal_dist are set as
        expected by MLDA"""
        _, model, _ = mv_simple()
        _, model_coarse, _ = mv_simple_coarse()
        with model:
            sampler = MLDA(coarse_models=[model_coarse], base_sampler="Metropolis")
            assert isinstance(sampler.proposal_dist, RecursiveDAProposal)
            assert sampler.base_proposal_dist is None
            assert isinstance(sampler.step_method_below.proposal_dist, NormalProposal)

            sampler = MLDA(coarse_models=[model_coarse])
            assert isinstance(sampler.proposal_dist, RecursiveDAProposal)
            assert sampler.base_proposal_dist is None
            assert isinstance(sampler.step_method_below.proposal_dist, UniformProposal)

            s = np.ones(model.ndim)
            sampler = MLDA(coarse_models=[model_coarse], base_sampler="Metropolis", base_S=s)
            assert isinstance(sampler.proposal_dist, RecursiveDAProposal)
            assert sampler.base_proposal_dist is None
            assert isinstance(sampler.step_method_below.proposal_dist, NormalProposal)

            sampler = MLDA(coarse_models=[model_coarse], base_S=s)
            assert isinstance(sampler.proposal_dist, RecursiveDAProposal)
            assert sampler.base_proposal_dist is None
            assert isinstance(sampler.step_method_below.proposal_dist, UniformProposal)

            s = np.diag(s)
            sampler = MLDA(coarse_models=[model_coarse], base_sampler="Metropolis", base_S=s)
            assert isinstance(sampler.proposal_dist, RecursiveDAProposal)
            assert sampler.base_proposal_dist is None
            assert isinstance(sampler.step_method_below.proposal_dist, MultivariateNormalProposal)

            sampler = MLDA(coarse_models=[model_coarse], base_S=s)
            assert isinstance(sampler.proposal_dist, RecursiveDAProposal)
            assert sampler.base_proposal_dist is None
            assert isinstance(sampler.step_method_below.proposal_dist, UniformProposal)

            s[0, 0] = -s[0, 0]
            with pytest.raises(np.linalg.LinAlgError):
                MLDA(coarse_models=[model_coarse], base_sampler="Metropolis", base_S=s)

    def test_step_methods_in_each_level(self):
        """Test that MLDA creates the correct hierarchy of step methods when no
        coarse models are passed and when two coarse models are passed."""
        _, model, _ = mv_simple()
        _, model_coarse, _ = mv_simple_coarse()
        _, model_very_coarse, _ = mv_simple_very_coarse()
        with model:
            s = np.ones(model.ndim) + 2.0
            sampler = MLDA(
                coarse_models=[model_very_coarse, model_coarse],
                base_S=s,
                base_sampler="Metropolis",
            )
            assert isinstance(sampler.step_method_below, MLDA)
            assert isinstance(sampler.step_method_below.step_method_below, Metropolis)
            assert np.all(sampler.step_method_below.step_method_below.proposal_dist.s == s)

            sampler = MLDA(coarse_models=[model_very_coarse, model_coarse], base_S=s)
            assert isinstance(sampler.step_method_below, MLDA)
            assert isinstance(sampler.step_method_below.step_method_below, DEMetropolisZ)
            assert np.all(sampler.step_method_below.step_method_below.proposal_dist.s == s)

    def test_exceptions_coarse_models(self):
        """Test that MLDA generates the expected exceptions when no coarse_models arg
        is passed, an empty list is passed or when coarse_models is not a list"""
        with pytest.raises(TypeError):
            _, model, _ = mv_simple()
            with model:
                MLDA()

        with pytest.raises(ValueError):
            _, model, _ = mv_simple()
            with model:
                MLDA(coarse_models=[])

        with pytest.raises(ValueError):
            _, model, _ = mv_simple()
            with model:
                MLDA(coarse_models=(model, model))

    def test_nonparallelized_chains_are_random(self):
        """Test that parallel chain are not identical when no parallelisation
        is applied"""
        with Model() as coarse_model:
            Normal("x", 0.3, 1)

        with Model():
            Normal("x", 0, 1)
            for stepper in TestMLDA.steppers:
                step = stepper(coarse_models=[coarse_model])
                trace = sample(chains=2, cores=1, draws=20, tune=0, step=step, random_seed=1)
                samples = np.array(trace.get_values("x", combine=False))[:, 5]
                assert (
                    len(set(samples)) == 2
                ), "Non parallelized {} " "chains are identical.".format(stepper)

    def test_parallelized_chains_are_random(self):
        """Test that parallel chain are
        not identical when parallelisation
        is applied"""
        with Model() as coarse_model:
            Normal("x", 0.3, 1)

        with Model():
            Normal("x", 0, 1)
            for stepper in TestMLDA.steppers:
                step = stepper(coarse_models=[coarse_model])
                trace = sample(chains=2, cores=2, draws=20, tune=0, step=step, random_seed=1)
                samples = np.array(trace.get_values("x", combine=False))[:, 5]
                assert len(set(samples)) == 2, "Parallelized {} " "chains are identical.".format(
                    stepper
                )

    def test_acceptance_rate_against_coarseness(self):
        """Test that the acceptance rate increases
        when the coarse model is closer to
        the fine model."""
        with Model() as coarse_model_0:
            Normal("x", 5.0, 1.0)

        with Model() as coarse_model_1:
            Normal("x", 6.0, 2.0)

        with Model() as coarse_model_2:
            Normal("x", 20.0, 5.0)

        possible_coarse_models = [coarse_model_0, coarse_model_1, coarse_model_2]
        acc = []

        with Model():
            Normal("x", 5.0, 1.0)
            for coarse_model in possible_coarse_models:
                step = MLDA(coarse_models=[coarse_model], subsampling_rates=3)
                trace = sample(chains=1, draws=500, tune=100, step=step, random_seed=1)
                acc.append(trace.get_sampler_stats("accepted").mean())
            assert acc[0] > acc[1] > acc[2], (
                "Acceptance rate is not "
                "strictly increasing when"
                "coarse model is closer to "
                "fine model. Acceptance rates"
                "were: {}".format(acc)
            )

    def test_mlda_non_blocked(self):
        """Test that MLDA correctly creates non-blocked
        compound steps in level 0 when using a Metropolis
        base sampler."""
        _, model = simple_2model_continuous()
        _, model_coarse = simple_2model_continuous()
        with model:
            for stepper in self.steppers:
                assert isinstance(
                    stepper(
                        coarse_models=[model_coarse],
                        base_sampler="Metropolis",
                        base_blocked=False,
                    ).step_method_below,
                    CompoundStep,
                )

    def test_mlda_blocked(self):
        """Test the type of base sampler instantiated
        when switching base_blocked flag while
        the base sampler is Metropolis and when
        the base sampler is DEMetropolisZ."""
        _, model = simple_2model_continuous()
        _, model_coarse = simple_2model_continuous()
        with model:
            for stepper in self.steppers:
                assert not isinstance(
                    stepper(
                        coarse_models=[model_coarse],
                        base_sampler="Metropolis",
                        base_blocked=True,
                    ).step_method_below,
                    CompoundStep,
                )
                assert isinstance(
                    stepper(
                        coarse_models=[model_coarse],
                        base_sampler="Metropolis",
                        base_blocked=True,
                    ).step_method_below,
                    Metropolis,
                )
                assert isinstance(
                    stepper(coarse_models=[model_coarse]).step_method_below,
                    DEMetropolisZ,
                )

    def test_tuning_and_scaling_on(self):
        """Test that tune and base_scaling change as expected when
        tuning is on."""
        np.random.seed(1234)
        ts = 100
        _, model = simple_2model_continuous()
        _, model_coarse = simple_2model_continuous()
        with model:
            trace_0 = sample(
                tune=ts,
                draws=20,
                step=MLDA(
                    coarse_models=[model_coarse],
                    base_sampler="Metropolis",
                    base_tune_interval=50,
                    base_scaling=100.0,
                ),
                chains=1,
                discard_tuned_samples=False,
                random_seed=1234,
            )

            trace_1 = sample(
                tune=ts,
                draws=20,
                step=MLDA(
                    coarse_models=[model_coarse],
                    base_tune_target="scaling",
                    base_tune_interval=50,
                    base_scaling=100.0,
                ),
                chains=1,
                discard_tuned_samples=False,
                random_seed=1234,
            )

            trace_2 = sample(
                tune=ts,
                draws=20,
                step=MLDA(coarse_models=[model_coarse], base_tune_interval=50, base_lamb=100.0),
                chains=1,
                discard_tuned_samples=False,
                random_seed=1234,
            )

        assert trace_0.get_sampler_stats("tune", chains=0)[0]
        assert trace_0.get_sampler_stats("tune", chains=0)[ts - 1]
        assert not trace_0.get_sampler_stats("tune", chains=0)[ts]
        assert not trace_0.get_sampler_stats("tune", chains=0)[-1]
        assert trace_0.get_sampler_stats("base_scaling", chains=0)[0, 0] == 100.0
        assert trace_0.get_sampler_stats("base_scaling", chains=0)[0, 1] == 100.0
        assert trace_0.get_sampler_stats("base_scaling", chains=0)[-1, 0] < 100.0
        assert trace_0.get_sampler_stats("base_scaling", chains=0)[-1, 1] < 100.0

        assert trace_1.get_sampler_stats("tune", chains=0)[0]
        assert trace_1.get_sampler_stats("tune", chains=0)[ts - 1]
        assert not trace_1.get_sampler_stats("tune", chains=0)[ts]
        assert not trace_1.get_sampler_stats("tune", chains=0)[-1]
        assert trace_1.get_sampler_stats("base_scaling", chains=0)[0] == 100.0
        assert trace_1.get_sampler_stats("base_scaling", chains=0)[-1] < 100.0

        assert trace_2.get_sampler_stats("tune", chains=0)[0]
        assert trace_2.get_sampler_stats("tune", chains=0)[ts - 1]
        assert not trace_2.get_sampler_stats("tune", chains=0)[ts]
        assert not trace_2.get_sampler_stats("tune", chains=0)[-1]
        assert trace_2.get_sampler_stats("base_lambda", chains=0)[0] == 100.0
        assert trace_2.get_sampler_stats("base_lambda", chains=0)[-1] < 100.0

    def test_tuning_and_scaling_off(self):
        """Test that tuning is deactivated when sample()'s tune=0 and that
        MLDA's tune=False is overridden by sample()'s tune."""
        np.random.seed(12345)
        _, model = simple_2model_continuous()
        _, model_coarse = simple_2model_continuous()

        ts_0 = 0
        with model:
            trace_0 = sample(
                tune=ts_0,
                draws=100,
                step=MLDA(
                    coarse_models=[model_coarse],
                    base_sampler="Metropolis",
                    base_tune_interval=50,
                    base_scaling=100.0,
                    tune=False,
                ),
                chains=1,
                discard_tuned_samples=False,
                random_seed=12345,
            )

        ts_1 = 100
        with model:
            trace_1 = sample(
                tune=ts_1,
                draws=20,
                step=MLDA(
                    coarse_models=[model_coarse],
                    base_sampler="Metropolis",
                    base_tune_interval=50,
                    base_scaling=100.0,
                    tune=False,
                ),
                chains=1,
                discard_tuned_samples=False,
                random_seed=12345,
            )

        assert not trace_0.get_sampler_stats("tune", chains=0)[0]
        assert not trace_0.get_sampler_stats("tune", chains=0)[-1]
        assert (
            trace_0.get_sampler_stats("base_scaling", chains=0)[0, 0]
            == trace_0.get_sampler_stats("base_scaling", chains=0)[-1, 0]
            == trace_0.get_sampler_stats("base_scaling", chains=0)[0, 1]
            == trace_0.get_sampler_stats("base_scaling", chains=0)[-1, 1]
            == 100.0
        )

        assert trace_1.get_sampler_stats("tune", chains=0)[0]
        assert trace_1.get_sampler_stats("tune", chains=0)[ts_1 - 1]
        assert not trace_1.get_sampler_stats("tune", chains=0)[ts_1]
        assert not trace_1.get_sampler_stats("tune", chains=0)[-1]
        assert trace_1.get_sampler_stats("base_scaling", chains=0)[0, 0] == 100.0
        assert trace_1.get_sampler_stats("base_scaling", chains=0)[0, 1] == 100.0
        assert trace_1.get_sampler_stats("base_scaling", chains=0)[-1, 0] < 100.0
        assert trace_1.get_sampler_stats("base_scaling", chains=0)[-1, 1] < 100.0

        ts_2 = 0
        with model:
            trace_2 = sample(
                tune=ts_2,
                draws=100,
                step=MLDA(
                    coarse_models=[model_coarse],
                    base_tune_interval=50,
                    base_lamb=100.0,
                    base_tune_target=None,
                ),
                chains=1,
                discard_tuned_samples=False,
                random_seed=12345,
            )

        assert not trace_2.get_sampler_stats("tune", chains=0)[0]
        assert not trace_2.get_sampler_stats("tune", chains=0)[-1]
        assert (
            trace_2.get_sampler_stats("base_lambda", chains=0)[0]
            == trace_2.get_sampler_stats("base_lambda", chains=0)[-1]
            == trace_2.get_sampler_stats("base_lambda", chains=0)[0]
            == trace_2.get_sampler_stats("base_lambda", chains=0)[-1]
            == 100.0
        )

    def test_trace_length(self):
        """Check if trace length is as expected."""
        tune = 100
        draws = 50
        with Model() as coarse_model:
            Normal("n", 0, 2.2, shape=(3,))
        with Model():
            Normal("n", 0, 2, shape=(3,))
            step = MLDA(coarse_models=[coarse_model])
            trace = sample(tune=tune, draws=draws, step=step, chains=1, discard_tuned_samples=False)
            assert len(trace) == tune + draws

    @pytest.mark.parametrize(
        "variable,has_grad,outcome",
        [("n", True, 1), ("n", False, 1), ("b", True, 0), ("b", False, 0)],
    )
    def test_competence(self, variable, has_grad, outcome):
        """Test if competence function returns expected
        results for different models"""
        with Model() as pmodel:
            Normal("n", 0, 2, shape=(3,))
            Binomial("b", n=2, p=0.3)
        assert MLDA.competence(pmodel[variable], has_grad=has_grad) == outcome

    def test_multiple_subsampling_rates(self):
        """Test that when you give a single integer it is applied to all levels and
        when you give a list the list is applied correctly."""
        with Model() as coarse_model_0:
            Normal("n", 0, 2.2, shape=(3,))
        with Model() as coarse_model_1:
            Normal("n", 0, 2.1, shape=(3,))
        with Model():
            Normal("n", 0, 2.0, shape=(3,))

            step_1 = MLDA(coarse_models=[coarse_model_0, coarse_model_1], subsampling_rates=3)
            assert len(step_1.subsampling_rates) == 2
            assert step_1.subsampling_rates[0] == step_1.subsampling_rates[1] == 3

            step_2 = MLDA(coarse_models=[coarse_model_0, coarse_model_1], subsampling_rates=[3, 4])
            assert step_2.subsampling_rates[0] == 3
            assert step_2.subsampling_rates[1] == 4

            with pytest.raises(ValueError):
                step_3 = MLDA(
                    coarse_models=[coarse_model_0, coarse_model_1],
                    subsampling_rates=[3, 4, 10],
                )

    def test_aem_mu_sigma(self):
        """Test that AEM estimates mu_B and Sigma_B in
        the coarse models of a 3-level LR example correctly"""
        # create data for linear regression
        if theano.config.floatX == "float32":
            p = "float32"
        else:
            p = "float64"
        np.random.seed(123456)
        size = 200
        true_intercept = 1
        true_slope = 2
        sigma = 1
        x = np.linspace(0, 1, size, dtype=p)
        # y = a + b*x
        true_regression_line = true_intercept + true_slope * x
        # add noise
        y = true_regression_line + np.random.normal(0, sigma ** 2, size)
        s = np.identity(y.shape[0], dtype=p)
        np.fill_diagonal(s, sigma ** 2)

        # forward model Op - here, just the regression equation
        class ForwardModel(Op):
            if theano.config.floatX == "float32":
                itypes = [tt.fvector]
                otypes = [tt.fvector]
            else:
                itypes = [tt.dvector]
                otypes = [tt.dvector]

            def __init__(self, x, pymc3_model):
                self.x = x
                self.pymc3_model = pymc3_model

            def perform(self, node, inputs, outputs):
                intercept = inputs[0][0]
                x_coeff = inputs[0][1]

                temp = intercept + x_coeff * x + self.pymc3_model.bias.get_value()
                with self.pymc3_model:
                    set_data({"model_output": temp})
                outputs[0][0] = np.array(temp)

        # create the coarse models with separate biases
        mout = []
        coarse_models = []

        with Model() as coarse_model_0:
            mu_B = Data("mu_B", np.zeros(y.shape, dtype=p))
            bias = Data("bias", 3.5 * np.ones(y.shape, dtype=p))
            Sigma_B = Data("Sigma_B", np.zeros((y.shape[0], y.shape[0]), dtype=p))
            model_output = Data("model_output", np.zeros(y.shape, dtype=p))
            Sigma_e = Data("Sigma_e", s)

            # Define priors
            intercept = Normal("Intercept", 0, sigma=20)
            x_coeff = Normal("x", 0, sigma=20)

            theta = tt.as_tensor_variable([intercept, x_coeff])

            mout.append(ForwardModel(x, coarse_model_0))

            # Define likelihood
            likelihood = MvNormal("y", mu=mout[0](theta) + mu_B, cov=Sigma_e, observed=y)

            coarse_models.append(coarse_model_0)

        with Model() as coarse_model_1:
            mu_B = Data("mu_B", np.zeros(y.shape, dtype=p))
            bias = Data("bias", 2.2 * np.ones(y.shape, dtype=p))
            Sigma_B = Data("Sigma_B", np.zeros((y.shape[0], y.shape[0]), dtype=p))
            model_output = Data("model_output", np.zeros(y.shape, dtype=p))
            Sigma_e = Data("Sigma_e", s)

            # Define priors
            intercept = Normal("Intercept", 0, sigma=20)
            x_coeff = Normal("x", 0, sigma=20)

            theta = tt.as_tensor_variable([intercept, x_coeff])

            mout.append(ForwardModel(x, coarse_model_1))

            # Define likelihood
            likelihood = MvNormal("y", mu=mout[1](theta) + mu_B, cov=Sigma_e, observed=y)

            coarse_models.append(coarse_model_1)

        # fine model and inference
        with Model() as model:
            bias = Data("bias", np.zeros(y.shape, dtype=p))
            model_output = Data("model_output", np.zeros(y.shape, dtype=p))
            Sigma_e = Data("Sigma_e", s)

            # Define priors
            intercept = Normal("Intercept", 0, sigma=20)
            x_coeff = Normal("x", 0, sigma=20)

            theta = tt.as_tensor_variable([intercept, x_coeff])

            mout.append(ForwardModel(x, model))

            # Define likelihood
            likelihood = MvNormal("y", mu=mout[-1](theta), cov=Sigma_e, observed=y)

            step_mlda = MLDA(coarse_models=coarse_models, adaptive_error_model=True)

            trace_mlda = sample(
                draws=100,
                step=step_mlda,
                chains=1,
                tune=200,
                discard_tuned_samples=True,
                random_seed=84759238,
            )

            m0 = step_mlda.step_method_below.model_below.mu_B.get_value()
            s0 = step_mlda.step_method_below.model_below.Sigma_B.get_value()
            m1 = step_mlda.model_below.mu_B.get_value()
            s1 = step_mlda.model_below.Sigma_B.get_value()

            assert np.all(np.abs(m0 + 3.5 * np.ones(y.shape, dtype=p)) < 1e-1)
            assert np.all(np.abs(m1 + 2.2 * np.ones(y.shape, dtype=p)) < 1e-1)
            assert np.all(np.abs(s0 < 1e-1))
            assert np.all(np.abs(s1 < 1e-1))

    def test_variance_reduction(self):
        """
        Test if the right stats are outputed when variance reduction is used in MLDA,
        if the output estimates are close (VR estimate vs. standard estimate from
        the first chain) and if the variance of VR is lower. Uses a linear regression
        model with multiple levels where approximate levels have fewer data.
        """
        # arithmetic precision
        if theano.config.floatX == "float32":
            p = "float32"
        else:
            p = "float64"

        # set up the model and data
        seed = 12345
        np.random.seed(seed)
        size = 100
        true_intercept = 1
        true_slope = 2
        sigma = 0.2
        x = np.linspace(0, 1, size, dtype=p)
        # y = a + b*x
        true_regression_line = true_intercept + true_slope * x
        # add noise
        y = true_regression_line + np.random.normal(0, sigma ** 2, size)
        s = sigma

        x_coarse_0 = x[::3]
        y_coarse_0 = y[::3]
        x_coarse_1 = x[::2]
        y_coarse_1 = y[::2]

        # MCMC parameters
        ndraws = 200
        ntune = 100
        nsub = 3
        nchains = 1

        # define likelihoods with different Q
        class Likelihood1(Op):
            if theano.config.floatX == "float32":
                itypes = [tt.fvector]
                otypes = [tt.fscalar]
            else:
                itypes = [tt.dvector]
                otypes = [tt.dscalar]

            def __init__(self, x, y, pymc3_model):
                self.x = x
                self.y = y
                self.pymc3_model = pymc3_model

            def perform(self, node, inputs, outputs):
                intercept = inputs[0][0]
                x_coeff = inputs[0][1]

                temp = np.array(intercept + x_coeff * self.x, dtype=p)
                self.pymc3_model.Q.set_value(np.array(x_coeff, dtype=p))
                outputs[0][0] = np.array(
                    -(0.5 / s ** 2) * np.sum((temp - self.y) ** 2, dtype=p), dtype=p
                )

        class Likelihood2(Op):
            if theano.config.floatX == "float32":
                itypes = [tt.fvector]
                otypes = [tt.fscalar]
            else:
                itypes = [tt.dvector]
                otypes = [tt.dscalar]

            def __init__(self, x, y, pymc3_model):
                self.x = x
                self.y = y
                self.pymc3_model = pymc3_model

            def perform(self, node, inputs, outputs):
                intercept = inputs[0][0]
                x_coeff = inputs[0][1]

                temp = np.array(intercept + x_coeff * self.x, dtype=p)
                self.pymc3_model.Q.set_value(temp.mean(dtype=p))
                outputs[0][0] = np.array(
                    -(0.5 / s ** 2) * np.sum((temp - self.y) ** 2, dtype=p), dtype=p
                )

        # run four MLDA steppers for all combinations of
        # base_sampler and forward model
        for stepper in ["Metropolis", "DEMetropolisZ"]:
            for f in [Likelihood1, Likelihood2]:
                mout = []
                coarse_models = []

                with Model() as coarse_model_0:
                    if theano.config.floatX == "float32":
                        Q = Data("Q", np.float32(0.0))
                    else:
                        Q = Data("Q", np.float64(0.0))

                    # Define priors
                    intercept = Normal("Intercept", 0, sigma=20)
                    x_coeff = Normal("x", 0, sigma=20)

                    theta = tt.as_tensor_variable([intercept, x_coeff])

                    mout.append(f(x_coarse_0, y_coarse_0, coarse_model_0))
                    Potential("likelihood", mout[0](theta))

                    coarse_models.append(coarse_model_0)

                with Model() as coarse_model_1:
                    if theano.config.floatX == "float32":
                        Q = Data("Q", np.float32(0.0))
                    else:
                        Q = Data("Q", np.float64(0.0))

                    # Define priors
                    intercept = Normal("Intercept", 0, sigma=20)
                    x_coeff = Normal("x", 0, sigma=20)

                    theta = tt.as_tensor_variable([intercept, x_coeff])

                    mout.append(f(x_coarse_1, y_coarse_1, coarse_model_1))
                    Potential("likelihood", mout[1](theta))

                    coarse_models.append(coarse_model_1)

                with Model() as model:
                    if theano.config.floatX == "float32":
                        Q = Data("Q", np.float32(0.0))
                    else:
                        Q = Data("Q", np.float64(0.0))

                    # Define priors
                    intercept = Normal("Intercept", 0, sigma=20)
                    x_coeff = Normal("x", 0, sigma=20)

                    theta = tt.as_tensor_variable([intercept, x_coeff])

                    mout.append(f(x, y, model))
                    Potential("likelihood", mout[-1](theta))

                    step = MLDA(
                        coarse_models=coarse_models,
                        base_sampler=stepper,
                        subsampling_rates=nsub,
                        variance_reduction=True,
                        store_Q_fine=True,
                    )

                    trace = sample(
                        draws=ndraws,
                        step=step,
                        chains=nchains,
                        tune=ntune,
                        cores=1,
                        discard_tuned_samples=True,
                        random_seed=seed,
                    )

                    # get fine level stats (standard method)
                    Q_2 = trace.get_sampler_stats("Q_2").reshape((nchains, ndraws))
                    Q_mean_standard = Q_2.mean(axis=1).mean()
                    Q_se_standard = np.sqrt(Q_2.var() / az.ess(np.array(Q_2, np.float64)))

                    # get VR stats
                    Q_mean_vr, Q_se_vr = extract_Q_estimate(trace, 3)

                    # compare standard and VR
                    assert isclose(Q_mean_standard, Q_mean_vr, rel_tol=1e-1)
                    assert Q_se_standard > Q_se_vr

                    # check consistency of QoI acroess levels.
                    if isinstance(f, Likelihood1):
                        Q_1_0 = np.concatenate(trace.get_sampler_stats("Q_1_0")).reshape(
                            (nchains, ndraws * nsub)
                        )
                        Q_2_1 = np.concatenate(trace.get_sampler_stats("Q_2_1")).reshape(
                            (nchains, ndraws)
                        )
                        assert Q_1_0.mean(axis=1) == 0.0
                        assert Q_2_1.mean(axis=1) == 0.0



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pickle
import unittest

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest
import theano
import theano.tensor as tt

import pymc3 as pm

from pymc3 import Deterministic, Potential
from pymc3.distributions import HalfCauchy, Normal, transforms
from pymc3.model import ValueGradFunction
from pymc3.tests.helpers import select_by_precision


class NewModel(pm.Model):
    def __init__(self, name="", model=None):
        super().__init__(name, model)
        assert pm.modelcontext(None) is self
        # 1) init variables with Var method
        self.Var("v1", pm.Normal.dist())
        self.v2 = pm.Normal("v2", mu=0, sigma=1)
        # 2) Potentials and Deterministic variables with method too
        # be sure that names will not overlap with other same models
        pm.Deterministic("d", tt.constant(1))
        pm.Potential("p", tt.constant(1))


class DocstringModel(pm.Model):
    def __init__(self, mean=0, sigma=1, name="", model=None):
        super().__init__(name, model)
        self.Var("v1", Normal.dist(mu=mean, sigma=sigma))
        Normal("v2", mu=mean, sigma=sigma)
        Normal("v3", mu=mean, sigma=HalfCauchy("sd", beta=10, testval=1.0))
        Deterministic("v3_sq", self.v3 ** 2)
        Potential("p1", tt.constant(1))


class TestBaseModel:
    def test_setattr_properly_works(self):
        with pm.Model() as model:
            pm.Normal("v1")
            assert len(model.vars) == 1
            with pm.Model("sub") as submodel:
                submodel.Var("v1", pm.Normal.dist())
                assert hasattr(submodel, "v1")
                assert len(submodel.vars) == 1
            assert len(model.vars) == 2
            with submodel:
                submodel.Var("v2", pm.Normal.dist())
                assert hasattr(submodel, "v2")
                assert len(submodel.vars) == 2
            assert len(model.vars) == 3

    def test_context_passes_vars_to_parent_model(self):
        with pm.Model() as model:
            assert pm.model.modelcontext(None) == model
            assert pm.Model.get_context() == model
            # a set of variables is created
            nm = NewModel()
            assert pm.Model.get_context() == model
            # another set of variables are created but with prefix 'another'
            usermodel2 = NewModel(name="another")
            assert pm.Model.get_context() == model
            assert usermodel2._parent == model
            # you can enter in a context with submodel
            with usermodel2:
                usermodel2.Var("v3", pm.Normal.dist())
                pm.Normal("v4")
                # this variable is created in parent model too
        assert "another_v2" in model.named_vars
        assert "another_v3" in model.named_vars
        assert "another_v3" in usermodel2.named_vars
        assert "another_v4" in model.named_vars
        assert "another_v4" in usermodel2.named_vars
        assert hasattr(usermodel2, "v3")
        assert hasattr(usermodel2, "v2")
        assert hasattr(usermodel2, "v4")
        # When you create a class based model you should follow some rules
        with model:
            m = NewModel("one_more")
        assert m.d is model["one_more_d"]
        assert m["d"] is model["one_more_d"]
        assert m["one_more_d"] is model["one_more_d"]


class TestNested:
    def test_nest_context_works(self):
        with pm.Model() as m:
            new = NewModel()
            with new:
                assert pm.modelcontext(None) is new
            assert pm.modelcontext(None) is m
        assert "v1" in m.named_vars
        assert "v2" in m.named_vars

    def test_named_context(self):
        with pm.Model() as m:
            NewModel(name="new")
        assert "new_v1" in m.named_vars
        assert "new_v2" in m.named_vars

    def test_docstring_example1(self):
        usage1 = DocstringModel()
        assert "v1" in usage1.named_vars
        assert "v2" in usage1.named_vars
        assert "v3" in usage1.named_vars
        assert "v3_sq" in usage1.named_vars
        assert len(usage1.potentials), 1

    def test_docstring_example2(self):
        with pm.Model() as model:
            DocstringModel(name="prefix")
        assert "prefix_v1" in model.named_vars
        assert "prefix_v2" in model.named_vars
        assert "prefix_v3" in model.named_vars
        assert "prefix_v3_sq" in model.named_vars
        assert len(model.potentials), 1

    def test_duplicates_detection(self):
        with pm.Model():
            DocstringModel(name="prefix")
            with pytest.raises(ValueError):
                DocstringModel(name="prefix")

    def test_model_root(self):
        with pm.Model() as model:
            assert model is model.root
            with pm.Model() as sub:
                assert model is sub.root


class TestObserved:
    def test_observed_rv_fail(self):
        with pytest.raises(TypeError):
            with pm.Model():
                x = Normal("x")
                Normal("n", observed=x)

    def test_observed_type(self):
        X_ = np.random.randn(100, 5)
        X = pm.floatX(theano.shared(X_))
        with pm.Model():
            x1 = pm.Normal("x1", observed=X_)
            x2 = pm.Normal("x2", observed=X)

        assert x1.type == X.type
        assert x2.type == X.type


class TestTheanoConfig:
    def test_set_testval_raise(self):
        with theano.config.change_flags(compute_test_value="off"):
            with pm.Model():
                assert theano.config.compute_test_value == "raise"
            assert theano.config.compute_test_value == "off"

    def test_nested(self):
        with theano.config.change_flags(compute_test_value="off"):
            with pm.Model(theano_config={"compute_test_value": "ignore"}):
                assert theano.config.compute_test_value == "ignore"
                with pm.Model(theano_config={"compute_test_value": "warn"}):
                    assert theano.config.compute_test_value == "warn"
                assert theano.config.compute_test_value == "ignore"
            assert theano.config.compute_test_value == "off"


def test_matrix_multiplication():
    # Check matrix multiplication works between RVs, transformed RVs,
    # Deterministics, and numpy arrays
    with pm.Model() as linear_model:
        matrix = pm.Normal("matrix", shape=(2, 2))
        transformed = pm.Gamma("transformed", alpha=2, beta=1, shape=2)
        rv_rv = pm.Deterministic("rv_rv", matrix @ transformed)
        np_rv = pm.Deterministic("np_rv", np.ones((2, 2)) @ transformed)
        rv_np = pm.Deterministic("rv_np", matrix @ np.ones(2))
        rv_det = pm.Deterministic("rv_det", matrix @ rv_rv)
        det_rv = pm.Deterministic("det_rv", rv_rv @ transformed)

        posterior = pm.sample(
            10,
            tune=0,
            compute_convergence_checks=False,
            progressbar=False,
            return_inferencedata=False,
        )
        decimal = select_by_precision(7, 5)
        for point in posterior.points():
            npt.assert_almost_equal(
                point["matrix"] @ point["transformed"],
                point["rv_rv"],
                decimal=decimal,
            )
            npt.assert_almost_equal(
                np.ones((2, 2)) @ point["transformed"],
                point["np_rv"],
                decimal=decimal,
            )
            npt.assert_almost_equal(
                point["matrix"] @ np.ones(2),
                point["rv_np"],
                decimal=decimal,
            )
            npt.assert_almost_equal(
                point["matrix"] @ point["rv_rv"],
                point["rv_det"],
                decimal=decimal,
            )
            npt.assert_almost_equal(
                point["rv_rv"] @ point["transformed"],
                point["det_rv"],
                decimal=decimal,
            )


def test_duplicate_vars():
    with pytest.raises(ValueError) as err:
        with pm.Model():
            pm.Normal("a")
            pm.Normal("a")
    err.match("already exists")

    with pytest.raises(ValueError) as err:
        with pm.Model():
            pm.Normal("a")
            pm.Normal("a", transform=transforms.log)
    err.match("already exists")

    with pytest.raises(ValueError) as err:
        with pm.Model():
            a = pm.Normal("a")
            pm.Potential("a", a ** 2)
    err.match("already exists")

    with pytest.raises(ValueError) as err:
        with pm.Model():
            pm.Binomial("a", 10, 0.5)
            pm.Normal("a", transform=transforms.log)
    err.match("already exists")


def test_empty_observed():
    data = pd.DataFrame(np.ones((2, 3)) / 3)
    data.values[:] = np.nan
    with pm.Model():
        a = pm.Normal("a", observed=data)
        npt.assert_allclose(a.tag.test_value, np.zeros((2, 3)))
        b = pm.Beta("b", alpha=1, beta=1, observed=data)
        npt.assert_allclose(b.tag.test_value, np.ones((2, 3)) / 2)


class TestValueGradFunction(unittest.TestCase):
    def test_no_extra(self):
        a = tt.vector("a")
        a.tag.test_value = np.zeros(3, dtype=a.dtype)
        a.dshape = (3,)
        a.dsize = 3
        f_grad = ValueGradFunction([a.sum()], [a], [], mode="FAST_COMPILE")
        assert f_grad.size == 3

    def test_invalid_type(self):
        a = tt.ivector("a")
        a.tag.test_value = np.zeros(3, dtype=a.dtype)
        a.dshape = (3,)
        a.dsize = 3
        with pytest.raises(TypeError) as err:
            ValueGradFunction([a.sum()], [a], [], mode="FAST_COMPILE")
        err.match("Invalid dtype")

    def setUp(self):
        extra1 = tt.iscalar("extra1")
        extra1_ = np.array(0, dtype=extra1.dtype)
        extra1.tag.test_value = extra1_
        extra1.dshape = tuple()
        extra1.dsize = 1

        val1 = tt.vector("val1")
        val1_ = np.zeros(3, dtype=val1.dtype)
        val1.tag.test_value = val1_
        val1.dshape = (3,)
        val1.dsize = 3

        val2 = tt.matrix("val2")
        val2_ = np.zeros((2, 3), dtype=val2.dtype)
        val2.tag.test_value = val2_
        val2.dshape = (2, 3)
        val2.dsize = 6

        self.val1, self.val1_ = val1, val1_
        self.val2, self.val2_ = val2, val2_
        self.extra1, self.extra1_ = extra1, extra1_

        self.cost = extra1 * val1.sum() + val2.sum()

        self.f_grad = ValueGradFunction([self.cost], [val1, val2], [extra1], mode="FAST_COMPILE")

    def test_extra_not_set(self):
        with pytest.raises(ValueError) as err:
            self.f_grad.get_extra_values()
        err.match("Extra values are not set")

        with pytest.raises(ValueError) as err:
            self.f_grad(np.zeros(self.f_grad.size, dtype=self.f_grad.dtype))
        err.match("Extra values are not set")

    def test_grad(self):
        self.f_grad.set_extra_values({"extra1": 5})
        array = np.ones(self.f_grad.size, dtype=self.f_grad.dtype)
        val, grad = self.f_grad(array)
        assert val == 21
        npt.assert_allclose(grad, [5, 5, 5, 1, 1, 1, 1, 1, 1])

    def test_bij(self):
        self.f_grad.set_extra_values({"extra1": 5})
        array = np.ones(self.f_grad.size, dtype=self.f_grad.dtype)
        point = self.f_grad.array_to_dict(array)
        assert len(point) == 2
        npt.assert_allclose(point["val1"], 1)
        npt.assert_allclose(point["val2"], 1)

        array2 = self.f_grad.dict_to_array(point)
        npt.assert_allclose(array2, array)
        point_ = self.f_grad.array_to_full_dict(array)
        assert len(point_) == 3
        assert point_["extra1"] == 5

    def test_edge_case(self):
        # Edge case discovered in #2948
        ndim = 3
        with pm.Model() as m:
            pm.Lognormal(
                "sigma", mu=np.zeros(ndim), tau=np.ones(ndim), shape=ndim
            )  # variance for the correlation matrix
            pm.HalfCauchy("nu", beta=10)
            step = pm.NUTS()

        func = step._logp_dlogp_func
        func.set_extra_values(m.test_point)
        q = func.dict_to_array(m.test_point)
        logp, dlogp = func(q)
        assert logp.size == 1
        assert dlogp.size == 4
        npt.assert_allclose(dlogp, 0.0, atol=1e-5)

    def test_tensor_type_conversion(self):
        # case described in #3122
        X = np.random.binomial(1, 0.5, 10)
        X[0] = -1  # masked a single value
        X = np.ma.masked_values(X, value=-1)
        with pm.Model() as m:
            x1 = pm.Uniform("x1", 0.0, 1.0)
            x2 = pm.Bernoulli("x2", x1, observed=X)

        gf = m.logp_dlogp_function()

        assert m["x2_missing"].type == gf._extra_vars_shared["x2_missing"].type

    def test_theano_switch_broadcast_edge_cases(self):
        # Tests against two subtle issues related to a previous bug in Theano where tt.switch would not
        # always broadcast tensors with single values https://github.com/pymc-devs/aesara/issues/270

        # Known issue 1: https://github.com/pymc-devs/pymc3/issues/4389
        data = np.zeros(10)
        with pm.Model() as m:
            p = pm.Beta("p", 1, 1)
            obs = pm.Bernoulli("obs", p=p, observed=data)
        # Assert logp is correct
        npt.assert_allclose(
            obs.logp(m.test_point),
            np.log(0.5) * 10,
        )

        # Known issue 2: https://github.com/pymc-devs/pymc3/issues/4417
        # fmt: off
        data = np.array([
            1.35202174, -0.83690274, 1.11175166, 1.29000367, 0.21282749,
            0.84430966, 0.24841369, 0.81803141, 0.20550244, -0.45016253,
        ])
        # fmt: on
        with pm.Model() as m:
            mu = pm.Normal("mu", 0, 5)
            obs = pm.TruncatedNormal("obs", mu=mu, sigma=1, lower=-1, upper=2, observed=data)
        # Assert dlogp is correct
        npt.assert_allclose(m.dlogp([mu])({"mu": 0}), 2.499424682024436, rtol=1e-5)


def test_multiple_observed_rv():
    "Test previously buggy MultiObservedRV comparison code."
    y1_data = np.random.randn(10)
    y2_data = np.random.randn(100)
    with pm.Model() as model:
        mu = pm.Normal("mu")
        x = pm.DensityDist(  # pylint: disable=unused-variable
            "x", pm.Normal.dist(mu, 1.0).logp, observed={"value": 0.1}
        )
    assert not model["x"] == model["mu"]
    assert model["x"] == model["x"]
    assert model["x"] in model.observed_RVs
    assert not model["x"] in model.vars


def test_tempered_logp_dlogp():
    with pm.Model() as model:
        pm.Normal("x")
        pm.Normal("y", observed=1)

    func = model.logp_dlogp_function()
    func.set_extra_values({})

    func_temp = model.logp_dlogp_function(tempered=True)
    func_temp.set_extra_values({})

    func_nograd = model.logp_dlogp_function(compute_grads=False)
    func_nograd.set_extra_values({})

    func_temp_nograd = model.logp_dlogp_function(tempered=True, compute_grads=False)
    func_temp_nograd.set_extra_values({})

    x = np.ones(func.size, dtype=func.dtype)
    assert func(x) == func_temp(x)
    assert func_nograd(x) == func(x)[0]
    assert func_temp_nograd(x) == func(x)[0]

    func_temp.set_weights(np.array([0.0], dtype=func.dtype))
    func_temp_nograd.set_weights(np.array([0.0], dtype=func.dtype))
    npt.assert_allclose(func(x)[0], 2 * func_temp(x)[0])
    npt.assert_allclose(func(x)[1], func_temp(x)[1])

    npt.assert_allclose(func_nograd(x), func(x)[0])
    npt.assert_allclose(func_temp_nograd(x), func_temp(x)[0])

    func_temp.set_weights(np.array([0.5], dtype=func.dtype))
    func_temp_nograd.set_weights(np.array([0.5], dtype=func.dtype))
    npt.assert_allclose(func(x)[0], 4 / 3 * func_temp(x)[0])
    npt.assert_allclose(func(x)[1], func_temp(x)[1])

    npt.assert_allclose(func_nograd(x), func(x)[0])
    npt.assert_allclose(func_temp_nograd(x), func_temp(x)[0])


def test_model_pickle(tmpdir):
    """Tests that PyMC3 models are pickleable"""
    with pm.Model() as model:
        x = pm.Normal("x")
        pm.Normal("y", observed=1)

    file_path = tmpdir.join("model.p")
    with open(file_path, "wb") as buff:
        pickle.dump(model, buff)


def test_model_pickle_deterministic(tmpdir):
    """Tests that PyMC3 models are pickleable"""
    with pm.Model() as model:
        x = pm.Normal("x")
        z = pm.Normal("z")
        pm.Deterministic("w", x / z)
        pm.Normal("y", observed=1)

    file_path = tmpdir.join("model.p")
    with open(file_path, "wb") as buff:
        pickle.dump(model, buff)



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import arviz as az
import numpy as np
import numpy.testing as npt
import theano.tensor as tt

from scipy import stats

import pymc3 as pm

from pymc3.tests.helpers import SeededTest
from pymc3.util import get_var_name


class KnownMean:
    def test_mean(self):
        for varname, expected in self.means.items():
            samples = self.samples[varname]
            npt.assert_allclose(expected, samples.mean(0), self.rtol, self.atol)


class KnownVariance:
    def test_var(self):
        for varname, expected in self.variances.items():
            samples = self.samples[varname]
            npt.assert_allclose(expected, samples.var(0), self.rtol, self.atol)


class KnownCDF:
    ks_thin = 5
    alpha = 0.001

    def test_kstest(self):
        for varname, cdf in self.cdfs.items():
            samples = self.samples[varname]
            if samples.ndim == 1:
                t, p = stats.kstest(samples[:: self.ks_thin], cdf=cdf)
                assert self.alpha < p
            elif samples.ndim == 2:
                pvals = []
                for samples_, cdf_ in zip(samples.T, cdf):
                    t, p = stats.kstest(samples_[:: self.ks_thin], cdf=cdf_)
                    pvals.append(p)
                t, p = stats.combine_pvalues(pvals)
                assert self.alpha < p
            else:
                raise NotImplementedError()


class UniformFixture(KnownMean, KnownVariance, KnownCDF):
    means = {"a": 0}
    variances = {"a": 1.0 / 3}
    cdfs = {"a": stats.uniform(-1, 2).cdf}

    @classmethod
    def make_model(cls):
        model = pm.Model()
        with model:
            a = pm.Uniform("a", lower=-1, upper=1)
        return model


class NormalFixture(KnownMean, KnownVariance, KnownCDF):
    means = {"a": 2 * np.ones(10)}
    variances = {"a": 3 * np.ones(10)}
    cdfs = {"a": [stats.norm(2, np.sqrt(3)).cdf for _ in range(10)]}

    @classmethod
    def make_model(cls):
        with pm.Model() as model:
            a = pm.Normal("a", mu=2, sigma=np.sqrt(3), shape=10)
        return model


class BetaBinomialFixture(KnownCDF):
    cdfs = {"p": [stats.beta(a, b).cdf for a, b in zip([1.5, 2.5, 10], [3.5, 10.5, 1])]}

    @classmethod
    def make_model(cls):
        with pm.Model() as model:
            p = pm.Beta("p", [0.5, 0.5, 1.0], [0.5, 0.5, 1.0], shape=3)
            pm.Binomial("y", p=p, n=[4, 12, 9], observed=[1, 2, 9])
        return model


class StudentTFixture(KnownMean, KnownCDF):
    means = {"a": 0}
    cdfs = {"a": stats.t(df=4).cdf}
    ks_thin = 10

    @classmethod
    def make_model(cls):
        with pm.Model() as model:
            a = pm.StudentT("a", nu=4, mu=0, sigma=1)
        return model


class LKJCholeskyCovFixture(KnownCDF):
    cdfs = {
        "log_stds": [stats.norm(loc=x, scale=x / 10.0).cdf for x in [1, 2, 3, 4, 5]],
        # The entries of the correlation matrix should follow
        # beta(eta - 1 + d/2, eta - 1 + d/2) on (-1, 1).
        # See https://arxiv.org/abs/1309.7268
        "corr_entries_unit": [stats.beta(3 - 1 + 2.5, 3 - 1 + 2.5).cdf for _ in range(10)],
    }

    @classmethod
    def make_model(cls):
        with pm.Model() as model:
            sd_mu = np.array([1, 2, 3, 4, 5])
            sd_dist = pm.Lognormal.dist(mu=sd_mu, sigma=sd_mu / 10.0, shape=5)
            chol_packed = pm.LKJCholeskyCov("chol_packed", eta=3, n=5, sd_dist=sd_dist)
            chol = pm.expand_packed_triangular(5, chol_packed, lower=True)
            cov = tt.dot(chol, chol.T)
            stds = tt.sqrt(tt.diag(cov))
            pm.Deterministic("log_stds", tt.log(stds))
            corr = cov / stds[None, :] / stds[:, None]
            corr_entries_unit = (corr[np.tril_indices(5, -1)] + 1) / 2
            pm.Deterministic("corr_entries_unit", corr_entries_unit)
        return model


class BaseSampler(SeededTest):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.model = cls.make_model()
        with cls.model:
            cls.step = cls.make_step()
            cls.trace = pm.sample(
                cls.n_samples,
                tune=cls.tune,
                step=cls.step,
                cores=cls.chains,
                return_inferencedata=False,
            )
        cls.samples = {}
        for var in cls.model.unobserved_RVs:
            cls.samples[get_var_name(var)] = cls.trace.get_values(var, burn=cls.burn)

    def test_neff(self):
        if hasattr(self, "min_n_eff"):
            n_eff = az.ess(self.trace[self.burn :])
            for var in n_eff:
                npt.assert_array_less(self.min_n_eff, n_eff[var])

    def test_Rhat(self):
        rhat = az.rhat(self.trace[self.burn :])
        for var in rhat:
            npt.assert_allclose(rhat[var], 1, rtol=0.01)


class NutsFixture(BaseSampler):
    @classmethod
    def make_step(cls):
        args = {}
        if hasattr(cls, "step_args"):
            args.update(cls.step_args)
        if "scaling" not in args:
            _, step = pm.sampling.init_nuts(n_init=10000, **args)
        else:
            step = pm.NUTS(**args)
        return step

    def test_target_accept(self):
        accept = self.trace[self.burn :]["mean_tree_accept"]
        npt.assert_allclose(accept.mean(), self.step.target_accept, 1)


class MetropolisFixture(BaseSampler):
    @classmethod
    def make_step(cls):
        args = {}
        if hasattr(cls, "step_args"):
            args.update(cls.step_args)
        return pm.Metropolis(**args)


class SliceFixture(BaseSampler):
    @classmethod
    def make_step(cls):
        args = {}
        if hasattr(cls, "step_args"):
            args.update(cls.step_args)
        return pm.Slice(**args)



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import multiprocessing
import os

import numpy as np
import pytest
import theano
import theano.tensor as tt

from theano.compile.ops import as_op

import pymc3 as pm
import pymc3.parallel_sampling as ps


def test_context():
    with pm.Model():
        pm.Normal("x")
        ctx = multiprocessing.get_context("spawn")
        pm.sample(tune=2, draws=2, chains=2, cores=2, mp_ctx=ctx, return_inferencedata=False)


class NoUnpickle:
    def __getstate__(self):
        return self.__dict__.copy()

    def __setstate__(self, state):
        raise AttributeError("This fails")


def test_bad_unpickle():
    with pm.Model() as model:
        pm.Normal("x")

    with model:
        step = pm.NUTS()
        step.no_unpickle = NoUnpickle()
        with pytest.raises(Exception) as exc_info:
            pm.sample(
                tune=2,
                draws=2,
                mp_ctx="spawn",
                step=step,
                cores=2,
                chains=2,
                compute_convergence_checks=False,
                return_inferencedata=False,
            )
        assert "could not be unpickled" in str(exc_info.getrepr(style="short"))


tt_vector = tt.TensorType(theano.config.floatX, [False])


@as_op([tt_vector, tt.iscalar], [tt_vector])
def _crash_remote_process(a, master_pid):
    if os.getpid() != master_pid:
        os.exit(0)
    return 2 * np.array(a)


def test_dill():
    with pm.Model():
        pm.Normal("x")
        pm.sample(
            tune=1,
            draws=1,
            chains=2,
            cores=2,
            pickle_backend="dill",
            mp_ctx="spawn",
            return_inferencedata=False,
        )


def test_remote_pipe_closed():
    master_pid = os.getpid()
    with pm.Model():
        x = pm.Normal("x", shape=2, mu=0.1)
        tt_pid = tt.as_tensor_variable(np.array(master_pid, dtype="int32"))
        pm.Normal("y", mu=_crash_remote_process(x, tt_pid), shape=2)

        step = pm.Metropolis()
        with pytest.raises(RuntimeError, match="Chain [0-9] failed"):
            pm.sample(
                step=step,
                mp_ctx="spawn",
                tune=2,
                draws=2,
                cores=2,
                chains=2,
                return_inferencedata=False,
            )


def test_abort():
    with pm.Model() as model:
        a = pm.Normal("a", shape=1)
        pm.HalfNormal("b")
        step1 = pm.NUTS([a])
        step2 = pm.Metropolis([model.b_log__])

    step = pm.CompoundStep([step1, step2])

    ctx = multiprocessing.get_context()
    proc = ps.ProcessAdapter(
        10,
        10,
        step,
        chain=3,
        seed=1,
        mp_ctx=ctx,
        start={"a": 1.0, "b_log__": 2.0},
        step_method_pickled=None,
        pickle_backend="pickle",
    )
    proc.start()
    proc.write_next()
    proc.abort()
    proc.join()


def test_explicit_sample():
    with pm.Model() as model:
        a = pm.Normal("a", shape=1)
        pm.HalfNormal("b")
        step1 = pm.NUTS([a])
        step2 = pm.Metropolis([model.b_log__])

    step = pm.CompoundStep([step1, step2])

    ctx = multiprocessing.get_context()
    proc = ps.ProcessAdapter(
        10,
        10,
        step,
        chain=3,
        seed=1,
        mp_ctx=ctx,
        start={"a": 1.0, "b_log__": 2.0},
        step_method_pickled=None,
        pickle_backend="pickle",
    )
    proc.start()
    while True:
        proc.write_next()
        out = ps.ProcessAdapter.recv_draw([proc])
        view = proc.shared_point_view
        for name in view:
            view[name].copy()
        if out[1]:
            break
    proc.join()


def test_iterator():
    with pm.Model() as model:
        a = pm.Normal("a", shape=1)
        pm.HalfNormal("b")
        step1 = pm.NUTS([a])
        step2 = pm.Metropolis([model.b_log__])

    step = pm.CompoundStep([step1, step2])

    start = {"a": 1.0, "b_log__": 2.0}
    sampler = ps.ParallelSampler(10, 10, 3, 2, [2, 3, 4], [start] * 3, step, 0, False)
    with sampler:
        for draw in sampler:
            pass


def test_spawn_densitydist_function():
    with pm.Model() as model:
        mu = pm.Normal("mu", 0, 1)

        def func(x):
            return -2 * (x ** 2).sum()

        obs = pm.DensityDist("density_dist", func, observed=np.random.randn(100))
        trace = pm.sample(
            draws=10,
            tune=10,
            step=pm.Metropolis(),
            cores=2,
            mp_ctx="spawn",
            return_inferencedata=False,
        )


def test_spawn_densitydist_bound_method():
    with pm.Model() as model:
        mu = pm.Normal("mu", 0, 1)
        normal_dist = pm.Normal.dist(mu, 1)
        obs = pm.DensityDist("density_dist", normal_dist.logp, observed=np.random.randn(100))
        msg = "logp for DensityDist is a bound method, leading to RecursionError while serializing"
        with pytest.raises(ValueError, match=msg):
            trace = pm.sample(
                draws=10,
                tune=10,
                step=pm.Metropolis(),
                cores=2,
                mp_ctx="spawn",
                return_inferencedata=False,
            )


def test_spawn_densitydist_syswarning(monkeypatch):
    monkeypatch.setattr("pymc3.distributions.distribution.PLATFORM", "win32")
    with pm.Model() as model:
        mu = pm.Normal("mu", 0, 1)
        normal_dist = pm.Normal.dist(mu, 1)
        with pytest.warns(UserWarning, match="errors when sampling on platforms"):
            obs = pm.DensityDist("density_dist", normal_dist.logp, observed=np.random.randn(100))


def test_spawn_densitydist_mpctxwarning(monkeypatch):
    ctx = multiprocessing.get_context("spawn")
    monkeypatch.setattr(multiprocessing, "get_context", lambda: ctx)
    with pm.Model() as model:
        mu = pm.Normal("mu", 0, 1)
        normal_dist = pm.Normal.dist(mu, 1)
        with pytest.warns(UserWarning, match="errors when sampling when multiprocessing"):
            obs = pm.DensityDist("density_dist", normal_dist.logp, observed=np.random.randn(100))



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numbers

from copy import copy

import numpy as np
import theano.tensor as tt

from pymc3 import distributions as pm_dists
from pymc3.model import modelcontext

__all__ = ["Normal", "StudentT", "Binomial", "Poisson", "NegativeBinomial"]

# Define link functions

# Hack as assigning a function in the class definition automatically binds
# it as a method.


class Identity:
    def __call__(self, x):
        return x


identity = Identity()
logit = tt.nnet.sigmoid
inverse = tt.inv
exp = tt.exp


class Family:
    """Base class for Family of likelihood distribution and link functions."""

    priors = {}
    link = None

    def __init__(self, **kwargs):
        # Overwrite defaults
        for key, val in kwargs.items():
            if key == "priors":
                self.priors = copy(self.priors)
                self.priors.update(val)
            else:
                setattr(self, key, val)

    def _get_priors(self, model=None, name=""):
        """Return prior distributions of the likelihood.

        Returns
        -------
        dict: mapping name -> pymc3 distribution
        """
        if name:
            name = f"{name}_"
        model = modelcontext(model)
        priors = {}
        for key, val in self.priors.items():
            if isinstance(val, (numbers.Number, np.ndarray, np.generic)):
                priors[key] = val
            else:
                priors[key] = model.Var(f"{name}{key}", val)

        return priors

    def create_likelihood(self, name, y_est, y_data, model=None):
        """Create likelihood distribution of observed data.

        Parameters
        ----------
        y_est: theano.tensor
            Estimate of dependent variable
        y_data: array
            Observed dependent variable
        """
        priors = self._get_priors(model=model, name=name)
        # Wrap y_est in link function
        priors[self.parent] = self.link(y_est)
        if name:
            name = f"{name}_"
        return self.likelihood(f"{name}y", observed=y_data, **priors)

    def __repr__(self):
        return """Family {klass}:
    Likelihood  : {likelihood}({parent})
    Priors      : {priors}
    Link function: {link}.""".format(
            klass=self.__class__,
            likelihood=self.likelihood.__name__,
            parent=self.parent,
            priors=self.priors,
            link=self.link,
        )


class StudentT(Family):
    link = identity
    likelihood = pm_dists.StudentT
    parent = "mu"
    priors = {"lam": pm_dists.HalfCauchy.dist(beta=10, testval=1.0), "nu": 1}


class Normal(Family):
    link = identity
    likelihood = pm_dists.Normal
    parent = "mu"
    priors = {"sd": pm_dists.HalfCauchy.dist(beta=10, testval=1.0)}


class Binomial(Family):
    link = logit
    likelihood = pm_dists.Binomial
    parent = "p"
    priors = {"n": 1}


class Poisson(Family):
    link = exp
    likelihood = pm_dists.Poisson
    parent = "mu"
    priors = {"mu": pm_dists.HalfCauchy.dist(beta=10, testval=1.0)}


class NegativeBinomial(Family):
    link = exp
    likelihood = pm_dists.NegativeBinomial
    parent = "mu"
    priors = {
        "mu": pm_dists.HalfCauchy.dist(beta=10, testval=1.0),
        "alpha": pm_dists.HalfCauchy.dist(beta=10, testval=1.0),
    }



#   Copyright 2020 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import theano
import theano.tensor as tt


def make_sens_ic(n_states, n_theta, floatX):
    r"""
    The sensitivity matrix will always have consistent form. (n_states, n_states + n_theta)

    If the first n_states entries of the parameters vector in the simulate call
    correspond to initial conditions of the system,
    then the first n_states columns of the sensitivity matrix should form
    an identity matrix.

    If the last n_theta entries of the parameters vector in the simulate call
    correspond to ode paramaters, then the last n_theta columns in
    the sensitivity matrix will be 0.

    Parameters
    ----------
    n_states : int
        Number of state variables in the ODE
    n_theta : int
        Number of ODE parameters
    floatX : str
        dtype to be used for the array

    Returns
    -------
    dydp : array
        1D-array of shape (n_states * (n_states + n_theta),), representing the initial condition of the sensitivities
    """

    # Initialize the sensitivity matrix to be 0 everywhere
    sens_matrix = np.zeros((n_states, n_states + n_theta), dtype=floatX)

    # Slip in the identity matrix in the appropirate place
    sens_matrix[:, :n_states] = np.eye(n_states, dtype=floatX)

    # We need the sensitivity matrix to be a vector (see augmented_function)
    # Ravel and return
    dydp = sens_matrix.ravel()
    return dydp


def augment_system(ode_func, n_states, n_theta):
    """
    Function to create augmented system.

    Take a function which specifies a set of differential equations and return
    a compiled function which allows for computation of gradients of the
    differential equation's solition with repsect to the parameters.

    Uses float64 even if floatX=float32, because the scipy integrator always uses float64.

    Parameters
    ----------
    ode_func: function
        Differential equation.  Returns array-like.
    n_states: int
        Number of rows of the sensitivity matrix. (n_states)
    n_theta: int
        Number of ODE parameters

    Returns
    -------
    system: function
        Augemted system of differential equations.
    """

    # Present state of the system
    t_y = tt.vector("y", dtype="float64")
    t_y.tag.test_value = np.ones((n_states,), dtype="float64")
    # Parameter(s).  Should be vector to allow for generaliztion to multiparameter
    # systems of ODEs.  Is m dimensional because it includes all initial conditions as well as ode parameters
    t_p = tt.vector("p", dtype="float64")
    t_p.tag.test_value = np.ones((n_states + n_theta,), dtype="float64")
    # Time.  Allow for non-automonous systems of ODEs to be analyzed
    t_t = tt.scalar("t", dtype="float64")
    t_t.tag.test_value = 2.459

    # Present state of the gradients:
    # Will always be 0 unless the parameter is the inital condition
    # Entry i,j is partial of y[i] wrt to p[j]
    dydp_vec = tt.vector("dydp", dtype="float64")
    dydp_vec.tag.test_value = make_sens_ic(n_states, n_theta, "float64")

    dydp = dydp_vec.reshape((n_states, n_states + n_theta))

    # Get symbolic representation of the ODEs by passing tensors for y, t and theta
    yhat = ode_func(t_y, t_t, t_p[n_states:])
    # Stack the results of the ode_func into a single tensor variable
    if not isinstance(yhat, (list, tuple)):
        yhat = (yhat,)
    t_yhat = tt.stack(yhat, axis=0)

    # Now compute gradients
    J = tt.jacobian(t_yhat, t_y)

    Jdfdy = tt.dot(J, dydp)

    grad_f = tt.jacobian(t_yhat, t_p)

    # This is the time derivative of dydp
    ddt_dydp = (Jdfdy + grad_f).flatten()

    system = theano.function(
        inputs=[t_y, t_t, t_p, dydp_vec], outputs=[t_yhat, ddt_dydp], on_unused_input="ignore"
    )

    return system
